import pytest

from measured import Quantity
from measured.si import Celsius, Kelvin
from measured.us import Fahrenheit, Rankine


# https://en.wikipedia.org/wiki/Rankine_scale
@pytest.mark.parametrize(
    "kelvin, celsius, fahrenheit, rankine",
    [
        # Absolute zero
        (0 * Kelvin, -273.15 * Celsius, -459.67 * Fahrenheit, 0 * Rankine),
        # freezing point of brine
        (
            255.372222222 * Kelvin,
            -17.777777777 * Celsius,
            0 * Fahrenheit,
            459.67 * Rankine,
        ),
        # freezing point of water
        (273.15 * Kelvin, 0 * Celsius, 32 * Fahrenheit, 491.67 * Rankine),
        # boiling point of water
        (
            373.1339 * Kelvin,
            99.9839 * Celsius,
            211.97102 * Fahrenheit,
            671.64102 * Rankine,
        ),
    ],
)
def test_temperature_conversion_scale(
    kelvin: Quantity, celsius: Quantity, fahrenheit: Quantity, rankine: Quantity
) -> None:
    kelvin.assert_approximates(celsius)
    celsius.assert_approximates(kelvin)
    celsius.assert_approximates(fahrenheit)
    fahrenheit.assert_approximates(celsius)
    fahrenheit.assert_approximates(rankine)
    rankine.assert_approximates(kelvin)
    rankine.assert_approximates(celsius)
    kelvin.assert_approximates(rankine)



import pytest

from measured import IdentityPrefix, One, Prefix
from measured.iec import Bit, Kibi, Mebi
from measured.si import Deci, Kilo, Mega, Meter, Micro, Milli, Second


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    if "prefix" in metafunc.fixturenames:
        examples = [Kibi, Mebi, Milli, Deci, Kilo, Mega]
        ids = [d.name for d in examples]
        metafunc.parametrize("prefix", examples, ids=ids)


def test_repr(prefix: Prefix) -> None:
    assert repr(prefix) == f"Prefix(base={prefix.base!r}, exponent={prefix.exponent!r})"


def test_repr_roundtrips(prefix: Prefix) -> None:
    assert eval(repr(prefix)) is prefix


def test_prefixes_scale_quantities_up() -> None:
    length = 3 * (Kilo * Meter)
    assert length == 3000 * Meter


def test_prefixes_scale_quantities_down() -> None:
    length = 3 * (Micro * Meter)
    assert length == 0.000003 * Meter


def test_equal_prefixes_cancel() -> None:
    speed = (15 * (Kilo * Meter)) / (3 * (Kilo * Second))
    assert speed == (5 * (Kilo * Meter)) / (Kilo * Second)
    assert speed == (5000 * Meter) / (1000 * Second)
    assert speed == 5 * Meter / Second


def test_different_prefixes_cancel() -> None:
    speed = (15 * (Mega * Meter)) / (3 * (Kilo * Second))
    assert speed == (15 * (Kilo * Meter)) / (3 * Second)
    assert speed == (5 * (Kilo * Meter)) / Second
    assert speed == 5000 * Meter / Second


def test_can_compare_quantities_with_different_prefixes() -> None:
    assert 1.0 * (Kibi * Bit) == 1.024 * (Kilo * Bit)


def test_can_multiply_prefixes_with_same_base() -> None:
    assert 5 * ((Kilo * Mega) * Meter) == 5000000000 * Meter


def test_can_multiply_prefixes_with_different_bases() -> None:
    iec_first = 5 * (Kibi * Mega) * Meter
    si_first = 5 * (Mega * Kibi) * Meter
    iec_first.assert_approximates(si_first, within=1e-3)
    iec_first.assert_approximates(5120000000 * Meter, within=1e-3)
    si_first.assert_approximates(5120000000 * Meter, within=1e-3)


def test_can_divide_prefixes_with_same_base() -> None:
    assert 5 * (Mega / Kilo) * Meter == 5000 * Meter


def test_can_divide_prefixes_with_different_bases() -> None:
    (5 * (Mebi / Kilo) * Meter).assert_approximates(5242.88 * Meter, within=1e-6)
    (5 * (Mega / Kibi) * Meter).assert_approximates(4882.8125 * Meter, within=1e-6)


def test_multiplying_produces_number_quantities() -> None:
    assert 5 * Kilo == 5000 * One
    assert 5 * Milli == 0.005 * One


def test_associative_with_multiplication() -> None:
    assert (5 * Kilo) * Meter == 5 * (Kilo * Meter)


def test_associative_in_denominator() -> None:
    assert 10000 * Meter / (5 * Kilo * Second) == 2 * Meter / Second


def test_multiplying_by_random_things() -> None:
    with pytest.raises(TypeError):
        "hello" * Kilo  # type: ignore

    with pytest.raises(TypeError):
        Kilo * "hello"  # type: ignore


def test_dividing_by_random_things() -> None:
    with pytest.raises(TypeError):
        "hello" / Kilo  # type: ignore

    with pytest.raises(TypeError):
        Kilo / "hello"  # type: ignore


def test_roots() -> None:
    assert Mega.root(0) == IdentityPrefix
    assert Mega.root(2) == Kilo
    assert Mebi.root(2) == Kibi


def test_only_integer_roots() -> None:
    with pytest.raises(TypeError):
        Deci.root(0.5)  # type: ignore


def test_whole_power_roots_only() -> None:
    with pytest.raises(ValueError):
        Deci.root(3)



"""
Defines the [International System of Units][1] ("SI") and [prefixes][2].

[1]: https://en.wikipedia.org/wiki/International_System_of_Units

[2]: https://en.wikipedia.org/wiki/Metric_prefix

Attributes: Base units

    Meter (Unit): The distance travelled by light in a vacuum in 1/299792458 seconds.

    Second (Unit): The duration of 9192631770 periods of the radiation corresponding to
        the transition between the two hyperfine levels of the ground state of the
        caesium-133 atom.

    Gram (Unit): The kilogram is defined by setting the Planck constant _h_ exactly
        to 6.62607015×10⁻³⁴ J⋅s (J = kg⋅m²⋅s⁻²), given the definitions of the metre
        and the second.  Note that `measured` uses `Gram` as the base unit, for
        simplicity.

    Ampere (Unit): The flow of exactly 1/1.602176634×10⁻¹⁹ times the elementary charge
        _e_ per second.

    Kelvin (Unit): The kelvin is defined by setting the fixed numerical value of the
        Boltzmann constant _k_ to 1.380649×10⁻²3 J⋅K⁻¹, (J = kg⋅m²⋅s⁻²), given the
        definition of the kilogram, the metre, and the second.

    Mole (Unit): The amount of substance of exactly 6.02214076×10²³ elementary entities.

    Candela (Unit): The luminous intensity, in a given direction, of a source that emits
        monochromatic radiation of frequency 5.4×10¹⁴ hertz and that has a radiant
        intensity in that direction of 1/683 watt per steradian.

Attributes: Derived units

    Hertz (Unit): measures `Frequency`

    Radian (Unit): measures `PlaneAngle`, a dimensionless `Number`

    Steradian (Unit): measures `SolidAngle`, a dimensionless `Number`

    Newton (Unit): measures `Force`
    Joule (Unit): measures `Energy`
    Watt (Unit): measures `Power`

    Coulomb (Unit): measures `Charge`
    Volt (Unit): measures `Potential`
    Farad (Unit): measures `Capacitance`
    Ohm (Unit): measures `Resistance`
    Siemens (Unit): measures `Conductance`
    Henry (Unit): measures `Inductance`

    Weber (Unit): measures `MagneticFlux`
    Tesla (Unit): measures `MagneticBField`

    Lumen (Unit): measures `LuminousFlux`
    Lux (Unit): measures `Illuminance`

    Becquerel (Unit): Radioactive decays per unit time (alias for `Hertz`)
    Gray (Unit): absorbed dose of ionizing radiation (measures `RadioactiveDose`)
    Sievert (Unit): equivalent dose of ionizing radiation (measures `RadioactiveDose`)

    Katal (Unit): measures `Catalysis`


Attributes: Prefixes (base 10)

    Yotta (Prefix): 10 ²⁴, symbol `Y`
    Zetta (Prefix): 10²¹, symbol `Z`
    Exa (Prefix): 10¹⁸, symbol `E`
    Peta (Prefix): 10¹⁵, symbol `P`
    Tera (Prefix): 10¹², symbol `T`
    Giga (Prefix): 10⁹, symbol `G`
    Mega (Prefix): 10⁶, symbol `M`
    Kilo (Prefix): 10³, symbol `k`
    Hecto (Prefix): 10 ², symbol `h`
    Deca (Prefix): 10¹, symbol `d`
    Deci (Prefix): 10⁻¹, symbol `d`
    Centi (Prefix): 10⁻², symbol `c`
    Milli (Prefix): 10⁻³, symbol `m`
    Micro (Prefix): 10⁻⁶, symbol `μ`
    Nano (Prefix): 10⁻⁹, symbol `n`
    Pico (Prefix): 10⁻¹², symbol `p`
    Femto (Prefix): 10⁻¹⁵, symbol `f`
    Atto (Prefix): 10⁻¹⁸, symbol `a`
    Zepto (Prefix): 10⁻²¹, symbol `z`
    Yocto (Prefix): 10⁻²⁴, symbol `y`

"""

from math import pi as π

from . import (
    AmountOfSubstance,
    Current,
    Length,
    LuminousIntensity,
    Mass,
    One,
    PlaneAngle,
    Prefix,
    Temperature,
    Time,
    Unit,
)

# https://en.wikipedia.org/wiki/Metric_prefix

Yotta = Prefix(10, 24, name="yotta", symbol="Y")
Zetta = Prefix(10, 21, name="zetta", symbol="Z")
Exa = Prefix(10, 18, name="exa", symbol="E")
Peta = Prefix(10, 15, name="peta", symbol="P")
Tera = Prefix(10, 12, name="tera", symbol="T")
Giga = Prefix(10, 9, name="giga", symbol="G")
Mega = Prefix(10, 6, name="mega", symbol="M")
Kilo = Prefix(10, 3, name="kilo", symbol="k")
Hecto = Prefix(10, 2, name="hecto", symbol="h")
Deca = Prefix(10, 1, name="deca", symbol="d")
Deci = Prefix(10, -1, name="deci", symbol="d")
Centi = Prefix(10, -2, name="centi", symbol="c")
Milli = Prefix(10, -3, name="milli", symbol="m")
Micro = Prefix(10, -6, name="micro", symbol="μ")
Nano = Prefix(10, -9, name="nano", symbol="n")
Pico = Prefix(10, -12, name="pico", symbol="p")
Femto = Prefix(10, -15, name="femto", symbol="f")
Atto = Prefix(10, -18, name="atto", symbol="a")
Zepto = Prefix(10, -21, name="zepto", symbol="z")
Yocto = Prefix(10, -24, name="yocto", symbol="y")


#
# https://en.wikipedia.org/wiki/International_System_of_Units
#


# Base Units
# https://en.wikipedia.org/wiki/SI_base_unit

Meter = Length.unit(name="meter", symbol="m")
Second = Time.unit(name="second", symbol="s")
Gram = Mass.unit(name="gram", symbol="g")
Ampere = Current.unit(name="ampere", symbol="A")
Kelvin = Temperature.unit(name="kelvin", symbol="K")
Mole = AmountOfSubstance.unit(name="mole", symbol="mol")
Candela = LuminousIntensity.unit(name="candela", symbol="cd")

# Technically, it's the kilogram that's the SI base unit, and the other derived units
# are derived in terms of the kilogram.
Kilogram = Mass.unit("kilogram", "kg")
Kilogram.equals(1000 * Gram)


# Derived Units
# https://en.wikipedia.org/wiki/SI_derived_unit

Hertz = Unit.derive(One / Second, name="hertz", symbol="Hz")

Radian = PlaneAngle.unit(name="radian", symbol="rad")
Steradian = PlaneAngle.unit(name="steradian", symbol="sr")

Newton = Unit.derive((Kilogram) * Meter / Second**2, name="newton", symbol="N")
Joule = Unit.derive(Meter * Newton, name="joule", symbol="J")
Watt = Unit.derive(Joule / Second, name="watt", symbol="W")

Coulomb = Unit.derive(Second * Ampere, name="coulomb", symbol="C")
Volt = Unit.derive(Watt / Ampere, name="volt", symbol="V")
Farad = Unit.derive(Coulomb / Volt, name="farad", symbol="F")
Ohm = Unit.derive(Volt / Ampere, name="ohm", symbol="Ω")
Siemens = Unit.derive(Ampere / Volt, name="siemens", symbol="S")
Henry = Unit.derive(Ohm * Second, name="henry", symbol="H")

Weber = Unit.derive(Joule / Ampere, name="weber", symbol="Wb")
Tesla = Unit.derive(Weber / Meter**2, name="tesla", symbol="T")

Celsius = Temperature.unit(name="celsius", symbol="°C")
Celsius.zero(273.15 * Kelvin)

Lumen = Unit.derive(Candela * Steradian, name="lumen", symbol="lm")
Lux = Unit.derive(Lumen / Meter**2, name="lux", symbol="lux")

Becquerel = Hertz
Gray = Unit.derive(Joule / (Kilo * Gram), name="gray", symbol="Gy")
Sievert = Unit.derive(Joule / (Kilo * Gram), name="sievert", symbol="Sv")

Katal = Unit.derive(Mole / Second, name="katal", symbol="kat")


# SI Accepted units
# https://en.wikipedia.org/wiki/Non-SI_units_mentioned_in_the_SI

Liter = Unit.derive((Deci * Meter) ** 3, name="liter", symbol="L")

Minute = Time.unit("minute", "min")
Minute.equals(60 * Second)

Hour = Time.unit("hour", "h")
Hour.equals(3600 * Second)
Hour.equals(60 * Minute)

# This is the SI-compatible "day" as a unit of measure, defined as 86,400 seconds.
# https://en.wikipedia.org/wiki/Day#International_System_of_Units_(SI)
Day = Time.unit("day", "d")
Day.equals(86400 * Second)
Day.equals(1440 * Minute)
Day.equals(24 * Hour)

Degree = PlaneAngle.unit("degree", "°")
Degree.equals(π / 180 * Radian)

Arcminute = PlaneAngle.unit("arcminute", "'")
Arcminute.equals(π / 10800 * Radian)
Arcminute.equals(1 / 60 * Degree)

Arcsecond = PlaneAngle.unit("arcsecond", '"')
Arcsecond.equals(π / (180 * 3600) * Radian)
Arcsecond.equals(1 / 3600 * Degree)
Arcsecond.equals(1 / 60 * Arcminute)



from measured import (
    Capacitance,
    Charge,
    Current,
    MagneticBField,
    MagneticFlux,
    Potential,
    Resistance,
)
from measured.si import (
    Ampere,
    Coulomb,
    Farad,
    Meter,
    Milli,
    Ohm,
    Second,
    Tesla,
    Volt,
    Weber,
)


def test_ampere() -> None:
    assert Ampere.name == "ampere"
    assert Ampere.symbol == "A"
    assert Ampere.dimension == Current


def test_coulomb() -> None:
    assert Coulomb.name == "coulomb"
    assert Coulomb.symbol == "C"
    assert Coulomb.dimension == Charge


def test_farad() -> None:
    assert Farad.name == "farad"
    assert Farad.symbol == "F"
    assert Farad.dimension == Capacitance


def test_ohm() -> None:
    assert Ohm.name == "ohm"
    assert Ohm.symbol == "Ω"
    assert Ohm.dimension == Resistance


def test_tesla() -> None:
    assert Tesla.name == "tesla"
    assert Tesla.symbol == "T"
    assert Tesla.dimension == MagneticBField


def test_volt() -> None:
    assert Volt.name == "volt"
    assert Volt.symbol == "V"
    assert Volt.dimension == Potential


def test_weber() -> None:
    assert Weber.name == "weber"
    assert Weber.symbol == "Wb"
    assert Weber.dimension == MagneticFlux


def test_ohms_law() -> None:
    I = 4 * Ampere
    V = 12 * Volt
    R = 3 * Ohm

    assert I == V / R
    assert R == V / I
    assert V == I * R


def test_capacitance_is_charge_per_potential() -> None:
    charge = 5 * Coulomb
    potential = 10 * Volt
    capacitance = charge / potential
    assert capacitance == 0.5 * Farad


def test_magnetic_inductance_is_magnetic_flux_per_area() -> None:
    flux = 12 * Volt * Second
    assert flux == 12 * Weber

    area = (1 * (Milli * Meter)) ** 2

    b_field = flux / area
    assert b_field == 12000000 * Tesla



from measured.si import Liter, Milli
from measured.us import (
    Cup,
    FluidOunce,
    Gallon,
    Hogshead,
    Mile,
    Rod,
    Tablespoon,
    Teaspoon,
)


def test_common_cooking_measures() -> None:
    assert 1 * Teaspoon == 4.92892159375 * Milli * Liter

    assert 1 * Tablespoon == 14.78676478125 * Milli * Liter
    assert 1 * Tablespoon == 3 * Teaspoon

    assert 1 * Cup == 236.5882365 * Milli * Liter
    assert 1 * Cup == 8 * FluidOunce


def test_abe() -> None:
    # https://www.reddit.com/r/theydidthemath/comments/1y38bs/calculation_of_abe_simpsons_gas_mileage_a_star_is/

    abes_car = (40 * Rod) / (1 * Hogshead)

    assert 40 * Rod == 0.125 * Mile
    assert 1 * Hogshead == 63 * Gallon

    assert abes_car == (0.125 / 63) * Mile / Gallon
    assert 0.00198 * Mile / Gallon <= abes_car <= 0.00199 * Mile / Gallon



import pytest

from measured import Quantity
from measured.si import Day, Hour, Minute, Second


@pytest.mark.parametrize(
    "days, hours, minutes, seconds",
    [
        (0 * Day, 0 * Hour, 0 * Minute, 0 * Second),
        (1 / 86400 * Day, 1 / 3600 * Hour, 1 / 60 * Minute, 1 * Second),
        (1 / 1440 * Day, 1 / 60 * Hour, 1 * Minute, 60 * Second),
        (1 / 24 * Day, 1 * Hour, 60 * Minute, 3600 * Second),
        (1 * Day, 24 * Hour, 1440 * Minute, 86400 * Second),
    ],
)
def test_time_conversions(
    days: Quantity, hours: Quantity, minutes: Quantity, seconds: Quantity
) -> None:
    assert days == hours
    assert days == minutes
    assert days == seconds

    hours.assert_approximates(days)
    assert hours == minutes
    assert hours == seconds

    minutes.assert_approximates(days)
    assert minutes == hours
    assert minutes == seconds

    seconds.assert_approximates(days)
    assert seconds == hours
    assert seconds == minutes



from fractions import Fraction

import pytest

from measured import Length, Number, One, Unit
from measured.si import Meter, Second


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    base = sorted(Unit.base(), key=lambda u: u.name or "")
    ids = [d.name for d in base]

    # just take a sample of units (with overlaps) to prevent the tests from
    # growing exponentially
    if "a" in metafunc.fixturenames:
        metafunc.parametrize("a", base[:10], ids=ids[:10])
    if "b" in metafunc.fixturenames:
        metafunc.parametrize("b", base[5:15], ids=ids[5:15])
    if "c" in metafunc.fixturenames:
        metafunc.parametrize("c", base[-10:], ids=ids[-10:])

    if "base" in metafunc.fixturenames:
        metafunc.parametrize("base", base, ids=ids)


@pytest.fixture
def identity() -> Unit:
    return One


def test_homogenous_under_addition(a: Unit, b: Unit) -> None:
    # https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity
    #
    # Only commensurable quantities (physical quantities having the same dimension) may
    # be compared, equated, added, or subtracted.
    if a is b:
        assert a + b == a
    else:
        with pytest.raises(TypeError):
            a + b


def test_homogenous_under_subtraction(a: Unit, b: Unit) -> None:
    # https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity
    #
    # Only commensurable quantities (physical quantities having the same dimension) may
    # be compared, equated, added, or subtracted.
    if a is b:
        assert a - b == a
    else:
        with pytest.raises(TypeError):
            a - b


def test_abelian_associativity(a: Unit, b: Unit, c: Unit) -> None:
    # https://en.wikipedia.org/wiki/Abelian_group
    assert (a * b) * c == a * (b * c)


def test_abelian_identity(identity: Unit, a: Unit) -> None:
    assert identity * a == a


def test_abelian_inverse(identity: Unit, a: Unit) -> None:
    inverse = a**-1
    assert inverse * a == a * inverse
    assert inverse * a == identity
    assert identity / a == inverse


def test_abelian_commutativity(a: Unit, b: Unit) -> None:
    assert a * b == b * a


def test_no_dimensional_exponentation(base: Unit) -> None:
    with pytest.raises(TypeError):
        base**base  # type: ignore


def test_no_floating_point_exponentation(base: Unit) -> None:
    with pytest.raises(TypeError):
        base**0.5  # type: ignore


def test_no_fractional_exponentation(base: Unit) -> None:
    with pytest.raises(TypeError):
        base ** Fraction(1, 2)  # type: ignore


def test_repr_roundtrips(base: Unit) -> None:
    from measured import Dimension, Prefix  # noqa: F401

    assert eval(repr(base)) is base


def test_base_units_factor_to_themselves(base: Unit) -> None:
    assert base.factors == {base: 1}


def test_one() -> None:
    assert isinstance(One, Unit)
    assert One.dimension is Number
    assert One.name == "one"
    assert One.symbol == "1"


def test_roots() -> None:
    assert One.root(0) == One
    assert One.root(2) == One
    assert One.root(3) == One
    assert One.root(100) == One
    assert Meter.root(0) == One
    assert (Meter**2).root(0) == One
    assert (Meter**2).root(2) == Meter
    assert (Meter**4).root(2) == Meter**2
    assert (Meter**16 / Second**12).root(4) == Meter**4 / Second**3


def test_only_integer_roots() -> None:
    with pytest.raises(TypeError):
        (Meter**2).root(0.5)  # type: ignore


def test_whole_power_roots_only() -> None:
    with pytest.raises(ValueError):
        (Meter**3).root(2)


def test_names_unique() -> None:
    with pytest.raises(ValueError, match="already defined"):
        Length.unit("meter", "totally unique")

    with pytest.raises(ValueError, match="already defined"):
        Unit.derive(Meter / Second, name="ohm", symbol="totally unique")


def test_symbols_unique() -> None:
    with pytest.raises(ValueError, match="already defined"):
        Length.unit("totally unique", "m")

    with pytest.raises(ValueError, match="already defined"):
        Unit.derive(Meter / Second, name="totally unique", symbol="Ω")



import pytest

from measured import One, Quantity
from measured.si import Hertz, Meter, Second


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    examples = [
        1 * Meter,
        2 * Second,
        3 * Hertz,
        4 * Meter / Second,
        5 * Meter**2,
        6 * Meter**2 / Second**2,
    ]

    for exemplar in ["a", "b", "c"]:
        if exemplar in metafunc.fixturenames:
            metafunc.parametrize(exemplar, examples)


@pytest.fixture
def identity() -> Quantity:
    return 1 * One


def test_multiplication_associativity(a: Quantity, b: Quantity, c: Quantity) -> None:
    assert (a * b) * c == a * (b * c)


def test_multiplication_identity(identity: Quantity, a: Quantity) -> None:
    assert a * identity == a
    assert identity * a == a


def test_multiplication_inverse(identity: Quantity, a: Quantity) -> None:
    inverse = a**-1
    assert inverse * a == a * inverse
    assert inverse * a == identity
    assert identity / a == inverse


def test_multiplication_commutativity(a: Quantity, b: Quantity) -> None:
    assert a * b == b * a



import json
import pickle
from copy import copy
from types import ModuleType
from typing import Generator, List, Union

try:
    import cloudpickle
except ImportError:  # pragma: no cover
    cloudpickle = None

import pytest

import measured.json
from measured import Area, Dimension, Length, Prefix, Quantity, Resistance, Unit
from measured.json import MeasuredJSONDecoder, MeasuredJSONEncoder, install, uninstall
from measured.si import Giga, Hertz, Kilo, Meter, Milli, Ohm


@pytest.fixture
def codecs_installed() -> Generator[None, None, None]:
    with measured.json.codecs_installed():
        yield


NamedType = Union[Dimension, Prefix, Unit]
MeasuredType = Union[NamedType, Quantity]

DIMENSIONS: List[NamedType] = [Length, Area, Resistance]
PREFIXES: List[NamedType] = [Kilo, Milli, Kilo * Giga]
UNITS: List[NamedType] = [Meter, Kilo * Meter, Hertz, Ohm]
NAMED = DIMENSIONS + PREFIXES + UNITS

QUANTITIES: List[MeasuredType] = [5 * Meter, 5 * Hertz, 5.1 * Ohm]

SERIALIZERS = [json, pickle] + ([cloudpickle] if cloudpickle else [])


@pytest.mark.parametrize("named", NAMED, ids=[d.name for d in NAMED])
def test_named_type_copies_are_singletons(named: NamedType) -> None:
    assert copy(named) is named


@pytest.mark.parametrize("quantity", QUANTITIES, ids=map(str, QUANTITIES))
def test_quantiy_copies_are_new(quantity: NamedType) -> None:
    assert copy(quantity) == quantity
    assert copy(quantity) is not quantity


@pytest.mark.parametrize("serializer", SERIALIZERS)
@pytest.mark.parametrize("named", NAMED, ids=[d.name for d in NAMED])
def test_named_type_serializer_roundtrip(
    codecs_installed: None,
    serializer: ModuleType,
    named: NamedType,
) -> None:
    prior_name = named.name
    prior_symbol = named.symbol

    roundtripped: NamedType = serializer.loads(serializer.dumps(named))

    assert roundtripped == named
    assert roundtripped is named
    assert roundtripped.name == prior_name
    assert roundtripped.symbol == prior_symbol


@pytest.mark.parametrize("serializer", SERIALIZERS)
@pytest.mark.parametrize("quantity", QUANTITIES, ids=map(str, QUANTITIES))
def test_quantity_serializer_roundtrip(
    codecs_installed: None,
    serializer: ModuleType,
    quantity: Quantity,
) -> None:
    roundtripped: Quantity = serializer.loads(serializer.dumps(quantity))

    assert roundtripped == quantity
    assert roundtripped is not quantity


@pytest.mark.parametrize("named", NAMED, ids=[d.name for d in NAMED])
def test_named_type_explicit_json_roundtrip(named: NamedType) -> None:
    prior_name = named.name
    prior_symbol = named.symbol

    roundtripped: NamedType = json.loads(
        json.dumps(named, cls=MeasuredJSONEncoder),
        cls=MeasuredJSONDecoder,
    )

    assert roundtripped == named
    assert roundtripped is named
    assert roundtripped.name == prior_name
    assert roundtripped.symbol == prior_symbol


INSTANCES = [Length, Kilo, Meter, 5 * Meter]
INSTANCE_IDS = [str(i) for i in INSTANCES]


@pytest.mark.parametrize("obj", INSTANCES, ids=INSTANCE_IDS)
def test_not_naturally_json_serializable(obj: MeasuredType) -> None:
    with pytest.raises(TypeError, match="is not JSON serializable"):
        json.dumps(obj)


@pytest.mark.parametrize("obj", INSTANCES, ids=INSTANCE_IDS)
def test_json_temporarily_installed(obj: MeasuredType) -> None:
    with pytest.raises(TypeError, match="is not JSON serializable"):
        json.dumps(obj)

    with measured.json.codecs_installed():
        assert json.loads(json.dumps(obj)) == obj

    with pytest.raises(TypeError, match="is not JSON serializable"):
        json.dumps(obj)


@pytest.mark.parametrize("obj", INSTANCES, ids=INSTANCE_IDS)
def test_json_installation(obj: MeasuredType) -> None:
    with pytest.raises(TypeError, match="is not JSON serializable"):
        json.dumps(obj)

    try:
        install()
        assert json.loads(json.dumps(obj)) == obj
    finally:
        uninstall()

    with pytest.raises(TypeError, match="is not JSON serializable"):
        json.dumps(obj)


def test_still_raises_on_unrecognized_types() -> None:
    with pytest.raises(TypeError, match="not JSON serializable"):
        json.dumps(object(), cls=MeasuredJSONEncoder)


def test_ignores_unknown_measured_types(codecs_installed: None) -> None:
    unknown = {"__measured__": "wat", "no": "idea"}
    assert json.loads(json.dumps(unknown)) == unknown



import pytest

from measured import Quantity
from measured.si import Giga, Gram, Kilo, Kilogram, Mega, Milli


@pytest.mark.parametrize(
    "grams, kilograms",
    [
        (1 * Gram, 0.001 * Kilogram),
        (1 * Kilo * Gram, 1 * Kilogram),
        (1 * (Kilo * Gram), 1 * Kilogram),
        (1000 * (Kilo * Gram), 1000 * Kilogram),
        (1000 * (Kilo * Gram), 1 * (Mega * Gram)),
        (1000 * (Kilo * Gram), 1 * Kilo * Kilogram),
        (1 * Milli * Kilogram, 1 * Gram),
        (1 * Mega * Kilogram, 1 * (Giga * Gram)),
    ],
)
def test_kilogram_and_gram_equivalent(grams: Quantity, kilograms: Quantity) -> None:
    """Give special consideration to the equivalence between grams and kilograms, as
    they are derived as separate units in order to simplify the energy-related units"""
    assert grams == kilograms
    assert kilograms == grams



# https://en.wikipedia.org/wiki/Abelian_group

import pytest

from measured import ConversionNotFound, Quantity
from measured.si import Hertz, Meter, Minute, Second
from measured.us import Foot, Inch


@pytest.mark.parametrize(
    "a, b, c",
    [
        (2 * Meter, 3 * Meter, 4 * Meter),
        (2 * Hertz, 3 * Hertz, 4 * Hertz),
        (2 * Meter / Second, 3 * Meter / Second, 4 * Meter / Second),
        (2 * Meter**2, 3 * Meter**2, 4 * Meter**2),
    ],
)
def test_abelian_addition_associativity(a: Quantity, b: Quantity, c: Quantity) -> None:
    assert (a + b) + c == a + (b + c)


@pytest.mark.parametrize(
    "a",
    [
        2 * Meter,
        3 * Hertz,
        4 * Meter / Second,
        5 * Meter**2,
    ],
)
def test_abelian_addition_identity(a: Quantity) -> None:
    identity = 0 * a.unit
    inverse = -a

    assert identity + a == a
    assert a + identity == a

    assert a - identity == a
    assert identity - a == inverse


@pytest.mark.parametrize(
    "a",
    [
        2 * Meter,
        3 * Hertz,
        4 * Meter / Second,
        5 * Meter**2,
    ],
)
def test_abelian_addition_inverse(a: Quantity) -> None:
    identity = 0 * a.unit
    inverse = -a

    assert inverse + a == a + inverse
    assert inverse + a == identity


@pytest.mark.parametrize(
    "a, b",
    [
        (2 * Meter, 3 * Meter),
        (2 * Hertz, 3 * Hertz),
        (2 * Meter / Second, 3 * Meter / Second),
        (2 * Meter**2, 3 * Meter**2),
    ],
)
def test_abelian_addition_commutativity(a: Quantity, b: Quantity) -> None:
    assert a + b == b + a


@pytest.mark.parametrize(
    "a, b, c",
    [
        (1 * Meter, 2 * Foot, 1.6096 * Meter),
        (1 * Meter, 2 * Inch, 1.0508 * Meter),
        (2 * Meter / Second, 30 * Foot / Minute, 2.1524 * Meter / Second),
    ],
)
def test_addition_of_different_units(a: Quantity, b: Quantity, c: Quantity) -> None:
    assert a + b == c


def test_addition_requires_same_dimension() -> None:
    with pytest.raises(ConversionNotFound):
        (1 * Meter) + (1 * Second)


@pytest.mark.parametrize(
    "a, b, c",
    [
        (1 * Meter, 2 * Foot, 0.39039999999999997 * Meter),
        (1 * Meter, 2 * Inch, 0.9492 * Meter),
        (2 * Meter / Second, 30 * Foot / Minute, 1.8476 * Meter / Second),
    ],
)
def test_subtraction_of_different_units(a: Quantity, b: Quantity, c: Quantity) -> None:
    assert a - b == c, str(a - b)


def test_subtraction_requires_same_dimension() -> None:
    with pytest.raises(ConversionNotFound):
        (1 * Meter) - (1 * Second)



from math import pi as π

import pytest

from measured import Number, One, PlaneAngle, Quantity, SolidAngle
from measured.si import Arcminute, Arcsecond, Degree, Radian, Steradian


def test_radian_is_dimensionless_but_unique() -> None:
    assert Radian.dimension is Number
    assert Radian.dimension is PlaneAngle
    assert Radian != One
    assert Radian != Steradian
    assert Radian.name == "radian"
    assert Radian.symbol == "rad"


def test_steradian_is_dimensionless_but_unique() -> None:
    assert Steradian.dimension is Number
    assert Steradian.dimension is SolidAngle
    assert Steradian != One
    assert Steradian != Radian
    assert Steradian.name == "steradian"
    assert Steradian.symbol == "sr"


@pytest.mark.parametrize(
    "left, right",
    [
        (0 * Radian, 0 * Degree),
        (π / 4 * Radian, 45 * Degree),
        (1 * Radian, 180 / π * Degree),
        (π / 2 * Radian, 90 * Degree),
        (2 * Radian, 360 / π * Degree),
        (π * Radian, 180 * Degree),
        (3 * π / 2 * Radian, 270 * Degree),
        (2 * π * Radian, 360 * Degree),
        (1 * Arcminute, 60 * Arcsecond),
        (1 * Degree, 60 * Arcminute),
        (1 * Arcsecond, 0.000004848136 * Radian),
        (1 * Arcminute, 0.000290888 * Radian),
    ],
)
def test_angle_conversions(left: Quantity, right: Quantity) -> None:
    left.assert_approximates(right)
    right.assert_approximates(left)



"""
The goal of the `measured` library is to provide a sound foundation for recording and
converting physical quantities, while maintaining the integrity of their units and
dimensions.

While it aims to be the fastest library of its kind, automatically tracking the units
and dimensions of quantities introduces significant overhead.  You can use `measured`
for applications where the accuracy of the units is more important than raw numerical
computing speed.

The value classes of `measured` should generally be treated as immutable.  Rather than
setting the `magnitude` of a [`Quantity`][measured.Quantity], consider instantiating a
new one, or performing arithmetic on an existing one to produce new values.  Instances
of [`Dimension`][measured.Dimension], [`Prefix`][measured.Prefix], and
[`Unit`][measured.Unit] are always singletons within a Python process.

Examples:

    >>> from measured.si import Meter, Second
    >>> distance = 10 * Meter
    >>> time = 2 * Second
    >>> speed = distance / time
    >>> assert speed == 5 * Meter / Second

Attributes: Fundamental dimensions

    Number (Dimension): the basis of [counting and measuring][1], used to define
        [dimensionless quantities][2]

        [1]: https://en.wikipedia.org/wiki/Number

        [2]: https://en.wikipedia.org/wiki/Dimensionless_quantity

    Length (Dimension): [distance][1] or extent through space

        [1]: https://en.wikipedia.org/wiki/Length

    Time (Dimension): [what clocks read][1], the intervals between events

        [1]: https://en.wikipedia.org/wiki/Time

    Mass (Dimension): how much [matter][1] is in a physical body

        [1]: https://en.wikipedia.org/wiki/Mass

    Current (Dimension): the net rate of [flow of electrical charge][1] through a
        conductor

        [1]: https://en.wikipedia.org/wiki/Electric_current

    Temperature (Dimension): the average amount of [kinetic energy][1] in a system

        [1]: https://en.wikipedia.org/wiki/Temperature

    AmountOfSubstance (Dimension): how many [elementary entities][1] are in an object

        [1]: https://en.wikipedia.org/wiki/Mole_(unit)

    LuminousIntensity (Dimension): how strong a [light source][1] is over a volume

        [1]: https://en.wikipedia.org/wiki/Luminous_intensity

    Information (Dimension): how much [entropy][1] is present in a random variable

        [1]: https://en.wikipedia.org/wiki/Information#Information_theory

Attributes: Derived dimensions

    Area (Dimension):
    Volume (Dimension):

    PlaneAngle (Dimension):
    SolidAngle (Dimension):

    Speed (Dimension):
    Acceleration (Dimension):
    Jerk (Dimension):
    Snap (Dimension):
    Crackle (Dimension):
    Pop (Dimension):

    Frequency (Dimension):

    Force (Dimension):
    Energy (Dimension):
    Power (Dimension):

    Charge (Dimension):
    Potential (Dimension):
    Capacitance (Dimension):
    Resistance (Dimension):
    Conductance (Dimension):
    Inductance (Dimension):

    MagneticFlux (Dimension):
    MagneticBField (Dimension):

    LuminousFlux (Dimension):
    Illuminance (Dimension):

    RadioactiveDose (Dimension):

    Catalysis (Dimension):



Attributes: Base Prefixes

    IdentityPrefix (Prefix): represents the number 1, expressed as the prefix 0⁰


Attributes: Base Units

    One (Unit): represents the number 1, expressed as a unit of dimension `Number`

"""

import sys
from collections import defaultdict
from functools import lru_cache, total_ordering
from importlib.metadata import version
from math import log
from typing import (
    Any,
    Callable,
    ClassVar,
    Dict,
    Generator,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
    overload,
)

from .formatting import superscript

if sys.version_info < (3, 9):  # pragma: no cover
    # math.gcd changed in Python 3.8 from a two-argument for to a variable argument form
    import math

    from typing_extensions import SupportsIndex

    def recursive_gcd(*integers: SupportsIndex) -> int:
        if len(integers) <= 2:
            return math.gcd(*integers)
        return math.gcd(integers[0], gcd(*integers[1:]))

    gcd = recursive_gcd

else:  # pragma: no cover
    from math import gcd


try:
    from icecream import ic
except ImportError:  # pragma: no cover
    ic = lambda *a: None if not a else (a[0] if len(a) == 1 else a)  # noqa


__version__ = version("measured")

NUMERIC_CLASSES = (int, float)
Numeric = Union[int, float]


class FractionalDimensionError(ValueError):
    """Raised when computing the nth root of a Dimension, Prefix, or Unit would result
    in a non-integer dimesion"""

    def __init__(self, degree: int, value: Union["Dimension", "Prefix", "Unit"]):
        super().__init__(
            f"Taking the root of degree {degree} of {value} would "
            "result in a fractional exponent"
        )


class Dimension:
    """Dimension represents the kind of physical quantity being measured.

    Unless you're doing something really cool, you probably won't instantiate new
    dimensions directly.  Instead, you'll import the base dimensions and combine them
    through multiplication, division, or exponentation.

    Attributes:
        name (Optional[str]): The name of this dimension, which may not be set in the
            case of complex dimensions

        symbol (Optional[str]): The conventional [dimensional
            analysis](https://en.wikipedia.org/wiki/Dimensional_analysis#Definition)
            symbol of this dimension

        exponents (Tuple[int, ...]): The exponents that define this dimension in terms
            of the fundamental dimensions

    Examples:

        >>> from measured import (
        ...     Number,
        ...     Length,
        ...     Time,
        ...     Mass,
        ...     Current,
        ...     Temperature,
        ...     AmountOfSubstance,
        ...     LuminousIntensity,
        ...     Information
        ... )
        >>> Length.symbol
        'L'
        >>> Length.name
        'length'

        These dimensions form the basis for more complex dimensions, which can be
        produced through multiplication, exponentation, and division.

        >>> from measured import Area, Volume, Frequency, Speed
        >>> assert Area == Length * Length
        >>> assert Volume == Length**3
        >>> assert Speed == Length / Time
        >>> assert Frequency == Number / Time

        `measured` attempts to maintain Dimensions as singletons, so they can be used in
        both equality and identity tests.

        >>> assert Volume is Length**3

        Dimensions are hashable and may be used as keys in dictionaries.

        >>> lookup = {Volume: 'spacious'}
        >>> lookup[Volume]
        'spacious'

        Dimensions can be serialized in a number of ways, preserving their identity.

        >>> import pickle
        >>> assert pickle.loads(pickle.dumps(Volume)) is Volume

        >>> import json
        >>> from measured.json import MeasuredJSONEncoder, MeasuredJSONDecoder
        >>> json.dumps(Length, cls=MeasuredJSONEncoder)
        '{"__measured__": "Dimension", "name": "length", "symbol": "L", ...}'

        While using `measured`'s [JSON codecs][measured.json], Dimensions may be
        deserialized directly from that JSON representation.

        >>> encoded = json.dumps(Length, cls=MeasuredJSONEncoder)
        >>> json.loads(encoded, cls=MeasuredJSONDecoder)
        Dimension(exponents=(0, 1, 0, 0, 0, 0, 0, 0, 0, 0), name='length', symbol='L')

        With measured's JSON codecs installed, you can omit passing the encoder and
        decoder.

        >>> from measured.json import codecs_installed
        >>> with codecs_installed():
        ...     assert json.loads(json.dumps(Volume)) is Volume
    """

    _known: ClassVar[Dict[Tuple[int, ...], "Dimension"]] = {}
    _initialized: bool = False

    _fundamental: ClassVar[List["Dimension"]] = []
    _by_name: ClassVar[Dict[str, "Dimension"]] = {}

    exponents: Tuple[int, ...]
    name: Optional[str]
    symbol: Optional[str]

    def __new__(
        cls,
        exponents: Tuple[int, ...],
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> "Dimension":
        if exponents in cls._known:
            return cls._known[exponents]

        self = super().__new__(cls)
        cls._known[exponents] = self
        return self

    def __init__(
        self,
        exponents: Tuple[int, ...],
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> None:
        if self._initialized:
            return
        self.exponents = exponents
        self.name = name
        self.symbol = symbol
        self._initialized = True

        if name:
            self._by_name[name] = self

    @classmethod
    def fundamental(cls) -> Iterable["Dimension"]:
        """Returns the registered fundamental dimensions

        Fundamental dimensions are not derived from other dimensions, and they will
        always have an `exponents` tuple with a single `1`, like `[0, 0, 1, 0, 0...]`
        (except for `Number`, which is all `0`s).

        The fundamental dimensions are those defined directly using
        [`Dimension.define`][measured.Dimension.define], like those provided by
        `measured`.  The length and order of these Dimensions is also the length and
        order of each Dimensions `exponents` tuple.
        """
        return list(cls._fundamental)

    @classmethod
    def define(cls, name: str, symbol: str) -> "Dimension":
        """Defines a new fundamental Dimension

        Used by `measured` itself to define the fundamental Dimensions.  You may define
        additional dimensions, but be aware that doing so will change the cardinality of
        the `exponents` tuple for _all_ defined Dimensions.
        """
        index = len(cls._fundamental)
        if index == 0:
            # the first dimension must be Number, with an exponent of zero (identity)
            exponents = tuple([0])
        else:
            # other dimensions start with their exponent at one, and all others zero
            exponents = tuple(([0] * index) + [1])

        dimension = cls(exponents, name=name, symbol=symbol)

        # resize the keys for all previously known dimensions to account
        # for this new fundamental dimension
        for previous in list(cls._known.values()):
            del cls._known[previous.exponents]
            previous.exponents += (0,)
            cls._known[previous.exponents] = previous

        cls._fundamental.append(dimension)

        return dimension

    @classmethod
    def derive(
        cls, dimension: "Dimension", name: str, symbol: Optional[str] = None
    ) -> "Dimension":
        """Registers a new named dimension derived from other dimension"""
        dimension.name = name
        dimension.symbol = symbol or str(dimension)
        cls._by_name[name] = dimension
        return dimension

    def unit(self, name: str, symbol: str) -> "Unit":
        """Define a new unit of this dimension"""
        return Unit.define(self, name, symbol)

    # Pickle support

    def __getnewargs_ex__(self) -> Tuple[Tuple[Tuple[int, ...]], Dict[str, Any]]:
        return (self.exponents,), {}

    # JSON support

    def __json__(self) -> Dict[str, Any]:
        return {
            "name": self.name,
            "symbol": self.symbol,
            "exponents": list(self.exponents),
        }

    @classmethod
    def __from_json__(cls, json_object: Dict[str, Any]) -> "Dimension":
        return Dimension(tuple(json_object["exponents"]))

    # Pydantic support

    @classmethod
    def __get_validators__(
        cls,
    ) -> Generator[Callable[[Union[str, "Dimension"]], "Dimension"], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, "Dimension"]) -> "Dimension":
        if isinstance(value, str):
            if value not in cls._by_name:
                raise ValueError(f"{value!r} is not a named Dimension")

            return cls._by_name[value]

        if isinstance(value, Dimension):
            return value

        raise ValueError(f"No conversion from {value!r} to Dimension")

    def __repr__(self) -> str:
        return (
            "Dimension("
            f"exponents={self.exponents!r}, "
            f"name={self.name!r}, symbol={self.symbol!r}"
            ")"
        )

    def __str__(self) -> str:
        if self.symbol:
            return self.symbol

        return (
            "⋅".join(
                f"{dimension.symbol}{superscript(self.exponents[i])}"
                for i, dimension in enumerate(self._fundamental)
                if self.exponents[i] != 0
            )
            or "?"
        )

    # https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity
    #
    # Only commensurable quantities (physical quantities having the same dimension)
    # may be compared, equated, added, or subtracted.

    def __add__(self, other: "Dimension") -> "Dimension":
        if self is not other:
            return NotImplemented

        return self

    def __sub__(self, other: "Dimension") -> "Dimension":
        if self is not other:
            return NotImplemented

        return self

    # https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity
    #
    # The dimensions form an [abelian
    # group](https://en.wikipedia.org/wiki/Abelian_group) under multiplication, so:
    # One may take ratios of incommensurable quantities (quantities with different
    # dimensions), and multiply or divide them.

    @staticmethod
    @lru_cache(maxsize=None)
    def _multiply(self: "Dimension", other: "Dimension") -> "Dimension":
        return Dimension(tuple(s + o for s, o in zip(self.exponents, other.exponents)))

    def __mul__(self, other: "Dimension") -> "Dimension":
        if not isinstance(other, Dimension):
            return NotImplemented

        return Dimension._multiply(self, other)

    @staticmethod
    @lru_cache(maxsize=None)
    def _divide(self: "Dimension", other: "Dimension") -> "Dimension":
        return Dimension(tuple(s - o for s, o in zip(self.exponents, other.exponents)))

    def __truediv__(self, other: "Dimension") -> "Dimension":
        if not isinstance(other, Dimension):
            return NotImplemented

        return Dimension._divide(self, other)

    def __pow__(self, power: int) -> "Dimension":
        if not isinstance(power, int):
            return NotImplemented

        return Dimension(tuple(s * power for s in self.exponents))

    def root(self, degree: int) -> "Dimension":
        """Returns the nth root of this Dimension"""
        if not isinstance(degree, int):
            raise TypeError(f"degree should be an integer, not {type(degree)}")

        if degree == 0:
            return Number

        if any(s // degree != s / degree for s in self.exponents):
            raise FractionalDimensionError(degree, self)

        return Dimension(tuple(s // degree for s in self.exponents))

    def as_ratio(self) -> Tuple["Dimension", "Dimension"]:
        """Returns this dimension, split into a numerator and denominator"""
        numerator = tuple(e if e >= 0 else 0 for e in self.exponents)
        denominator = tuple(-e if e < 0 else 0 for e in self.exponents)
        return Dimension(numerator), Dimension(denominator)


class Prefix:
    """Prefixes scale a [`Unit`][measured.Unit] up or down by a constant factor.

    Prefixes are defined in systems of factors with a common integer base, and
    successive positive and/or negative exponents, like the SI system of prefixes using
    base 10 (micro, milli, kilo, mega, etc) or the IEC system of binary prefixes (kibi,
    mebi, gibi, etc).

    Attributes:

        base (int): The base of the system, like `10` for the SI system
        exponent (int | float): The exponent that the base is raised to

    Examples:

        Prefixes have a base and an exponent:

        >>> from measured.si import Milli, Kilo
        >>> Milli.symbol, Milli.base, Milli.exponent
        ('m', 10, -3)
        >>> Kilo.symbol, Kilo.base, Kilo.exponent
        ('k', 10, 3)

        Prefixes can be multiplied by numbers, Prefixes, and Units:

        >>> from measured.si import Kilo, Mega, Meter
        >>> assert 1 * Kilo * Meter == 1000 * Meter
        >>> assert 1 * (Kilo * Kilo * Meter) == 1 * Mega * Meter

        Prefixes can be divided by Prefixes and other prefixed units:

        >>> from measured.si import Kilo, Mega, Meter, Second
        >>> assert (1 * Mega * Meter) / (1 * Kilo * Second) == 1000 * Meter / Second

        Prefixes can be raised to an exponent:

        >>> from measured.si import Kilo, Mega, Meter
        >>> assert (2 * Kilo * Meter)**2 == 4 * Mega * Meter**2

        Prefixes can translate between bases, although be careful with the loss of
        precision during this conversion:

        >>> from measured.si import Kilo
        >>> from measured.iec import Mebi, Bit
        >>> assert 1 * Mebi * Bit == 1048.576 * Kilo * Bit

    """

    _known: ClassVar[Dict[Tuple[int, Numeric], "Prefix"]] = {}
    _initialized: bool = False

    _by_name: ClassVar[Dict[str, "Prefix"]] = {}

    base: int
    exponent: Numeric
    name: Optional[str]
    symbol: Optional[str]

    def __new__(
        cls,
        base: int,
        exponent: Numeric,
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> "Prefix":
        key = (base, exponent)
        if key in cls._known:
            return cls._known[key]

        self = super().__new__(cls)
        cls._known[key] = self
        return self

    def __init__(
        self,
        base: int,
        exponent: Numeric,
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> None:
        if self._initialized:
            return

        self.base = base
        self.exponent = exponent
        self.name = name
        self.symbol = symbol
        self._initialized = True

        if name:
            self._by_name[name] = self

    # Pickle support

    def __getnewargs_ex__(self) -> Tuple[Tuple[int, Numeric], Dict[str, Any]]:
        return (self.base, self.exponent), {}

    # JSON support

    def __json__(self) -> Dict[str, Any]:
        return {
            "base": self.base,
            "exponent": self.exponent,
            "name": self.name,
            "symbol": self.symbol,
        }

    @classmethod
    def __from_json__(cls, json_object: Dict[str, Any]) -> "Prefix":
        return Prefix(json_object["base"], json_object["exponent"])

    # Pydantic support

    @classmethod
    def __get_validators__(
        cls,
    ) -> Generator[Callable[[Union[str, "Prefix"]], "Prefix"], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, "Prefix"]) -> "Prefix":
        if isinstance(value, str):
            if value not in cls._by_name:
                raise ValueError(f"{value!r} is not a named Prefix")

            return cls._by_name[value]

        if isinstance(value, Prefix):
            return value

        raise ValueError(f"No conversion from {value!r} to Prefix")

    def __repr__(self) -> str:
        return f"Prefix(base={self.base!r}, exponent={self.exponent!r})"

    def __str__(self) -> str:
        if self.symbol:
            return self.symbol
        if self.exponent == 0:
            return ""
        return f"{self.base}{superscript(self.exponent)}"

    def quantify(self) -> Numeric:
        return cast(Numeric, self.base**self.exponent)

    @overload
    def __mul__(self, other: "Prefix") -> "Prefix":
        ...  # pragma: no cover

    @overload
    def __mul__(self, other: "Unit") -> "Unit":
        ...  # pragma: no cover

    @overload
    def __mul__(self, other: Numeric) -> "Quantity":
        ...  # pragma: no cover

    def __mul__(
        self, other: Union["Prefix", "Unit", Numeric]
    ) -> Union["Prefix", "Unit", "Quantity"]:
        if isinstance(other, Prefix):
            if other.base == 0:
                return self
            elif self.base == 0:
                return other
            elif other.base == self.base:
                return Prefix(self.base, self.exponent + other.exponent)

            base, exponent = self.base, self.exponent
            exponent += other.exponent * (log(other.base) / log(self.base))

            return Prefix(base, exponent)

        if isinstance(other, Unit):
            return other.scale(self)

        if isinstance(other, NUMERIC_CLASSES):
            return (other * One) * self.quantify()

        return NotImplemented

    __rmul__ = __mul__

    def __truediv__(self, other: "Prefix") -> "Prefix":
        if not isinstance(other, Prefix):
            return NotImplemented

        if other.base == 0:
            return self
        elif self.base == 0:
            return Prefix(other.base, -other.exponent)
        elif other.base == self.base:
            return Prefix(self.base, self.exponent - other.exponent)

        base, exponent = self.base, self.exponent
        exponent -= other.exponent * (log(other.base) / log(self.base))

        return Prefix(base, exponent)

    def __pow__(self, power: int) -> "Prefix":
        return Prefix(self.base, self.exponent * power)

    def root(self, degree: int) -> "Prefix":
        """Returns the nth root of this Prefix"""
        if not isinstance(degree, int):
            raise TypeError(f"degree should be an integer, not {type(degree)}")

        if degree == 0:
            return IdentityPrefix

        if self.exponent // degree != self.exponent / degree:
            raise FractionalDimensionError(degree, self)

        return Prefix(self.base, int(self.exponent // degree))


class Unit:
    """Unit is a predetermined reference amount or definition for a measurable quantity

    `measured` includes a number of well-known units, and additional contributions are
    always welcome.  The SI system of units is covered in the [`measured.si`][] package.

    Attributes:

        name (Optional[str]): the name of this unit, which may not be set in the case
            of complex derived units

        symbol (Optional[str]): the symbol of this unit, which may not be set in the
            case of complex derived units

        prefix (Prefix): the prefix to multiply this unit by (`IdentityPrefix` for
            unprefixed units)

        factors (Mapping["Unit", int]): a mapping of the base units and their powers
            that make up a compound unit, or `{self: 1}` for base units

    Examples:

        >>> from measured import Length, Time, Speed
        >>> from measured.si import Meter, Second
        >>> Meter.name, Meter.symbol
        ('meter', 'm')

        >>> meter_per_second = Meter / Second
        >>> assert meter_per_second.dimension is Speed
        >>> assert meter_per_second.dimension is Length / Time
        >>> meter_per_second.name, meter_per_second.symbol
        (None, None)

        >>> back_to_meter = (meter_per_second * Second)
        >>> back_to_meter.name, back_to_meter.symbol
        ('meter', 'm')

        >>> assert Meter.factors == {Meter: 1}
        >>> assert Second.factors == {Second: 1}
        >>> assert meter_per_second.factors == {Meter: 1, Second: -1}
    """

    UnitKey = Tuple[Prefix, Tuple[Tuple["Unit", int], ...]]

    _known: ClassVar[Dict[UnitKey, "Unit"]] = {}
    _initialized: bool = False

    _base: ClassVar[Set["Unit"]] = set()
    _by_name: ClassVar[Dict[str, "Unit"]] = {}
    _by_symbol: ClassVar[Dict[str, "Unit"]] = {}

    prefix: Prefix
    factors: Mapping["Unit", int]
    name: Optional[str]
    symbol: Optional[str]

    def __new__(
        cls,
        prefix: Prefix,
        factors: Mapping["Unit", int],
        dimension: Dimension,
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> "Unit":
        key = cls._build_key(prefix, factors)
        if key in cls._known:
            return cls._known[key]

        if name and name in cls._by_name:
            return cls._by_name[name]

        self = super().__new__(cls)
        if not factors:
            key = cls._build_key(prefix, {self: 1})
        cls._known[key] = self
        return self

    def __init__(
        self,
        prefix: Prefix,
        factors: Mapping["Unit", int],
        dimension: Dimension,
        name: Optional[str] = None,
        symbol: Optional[str] = None,
    ) -> None:
        if self._initialized:
            return
        self.prefix = prefix
        self.factors = factors or {self: 1}
        self.dimension = dimension
        self.name = name
        self.symbol = symbol
        self.names = [name]
        self.symbols = [symbol]
        self._initialized = True

        if name:
            self._by_name[name] = self
        if symbol:
            self._by_symbol[symbol] = self

    @classmethod
    def _build_key(cls, prefix: Prefix, factors: Mapping["Unit", int]) -> UnitKey:
        factor_key = tuple(sorted(factors.items(), key=lambda pair: id(pair[0])))
        return (prefix, factor_key)

    @classmethod
    def base(cls) -> Iterable["Unit"]:
        return list(cls._base)

    @classmethod
    def define(cls, dimension: Dimension, name: str, symbol: str) -> "Unit":
        """Defines a new base unit"""
        if name and name in cls._by_name:
            raise ValueError(f"A unit named {name} is already defined")
        if symbol and symbol in cls._by_symbol:
            raise ValueError(f"A unit with symbol {symbol} is already defined")

        unit = cls(IdentityPrefix, {}, dimension, name, symbol)
        cls._base.add(unit)
        return unit

    @classmethod
    def derive(cls, unit: "Unit", name: str, symbol: str) -> "Unit":
        """Registers a new named unit derived from other units"""
        if name in cls._by_name and cls._by_name[name] is not unit:
            raise ValueError(f"A unit named {name} is already defined")
        if symbol in cls._by_symbol and cls._by_symbol[symbol] is not unit:
            raise ValueError(f"A unit with symbol {symbol} is already defined")

        if unit.name:
            unit.names.append(name)
            unit.symbols.append(symbol)
            return unit
        else:
            unit.name = name
            unit.symbol = symbol

        cls._by_name[name] = unit
        cls._by_symbol[symbol] = unit

        return unit

    @classmethod
    def named(cls, name: str) -> "Unit":
        """Returns the unit with the given name"""
        return cls._by_name[name]

    def equals(self, other: "Quantity") -> None:
        """Defines a conversion between this Unit and another"""
        if other.unit == self and self is not One:
            raise ValueError("No need to define conversions for a unit and itself")
        conversions.equate(1 * self, other)

    def zero(self, zero: "Quantity") -> None:
        """Defines this unit as a scale with a zero point at another Quantity"""
        if zero.unit == self:
            raise ValueError("No need to define conversions for a unit and itself")
        conversions.translate(self, zero)

    # Pickle support

    def __getnewargs_ex__(
        self,
    ) -> Tuple[Tuple[Prefix, Mapping["Unit", int], Dimension], Dict[str, Any]]:
        factors = {} if self.factors == {self: 1} else self.factors
        args = (self.prefix, factors, self.dimension)
        kwargs = {"name": self.name, "symbol": self.symbol}
        return args, kwargs

    # JSON support

    def __json__(self) -> Dict[str, Any]:
        prefix, factors = self._build_key(self.prefix, self.factors)
        return {
            "name": self.name,
            "symbol": self.symbol,
            "dimension": self.dimension,
            "prefix": prefix if prefix.quantify() != 1 else None,
            "factors": None if factors == ((self, 1),) else factors,
        }

    @classmethod
    def __from_json__(cls, json_object: Dict[str, Any]) -> "Unit":
        if not json_object["factors"]:
            return cls._by_name[json_object["name"]]
        prefix = json_object["prefix"] or Prefix(0, 0)
        factors = dict(json_object["factors"])
        dimension = json_object["dimension"]
        return Unit(prefix, factors, dimension)

    # Pydantic support

    @classmethod
    def __get_validators__(
        cls,
    ) -> Generator[Callable[[Union[str, "Unit"]], "Unit"], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, "Unit"]) -> "Unit":
        if isinstance(value, str):
            if value not in cls._by_name:
                raise ValueError(f"{value!r} is not a named Unit")

            return cls._by_name[value]

        if isinstance(value, Unit):
            return value

        raise ValueError(f"No conversion from {value!r} to Unit")

    def scale(self, prefix: Prefix) -> "Unit":
        """Given a Prefix, creates a new unit scaled by that Prefix"""
        return self.__class__(self.prefix * prefix, self.factors, self.dimension)

    @lru_cache(maxsize=None)
    def quantify(self) -> "Quantity":
        """Produce a Quantity of this Unit, including the Prefix.

        Examples:

            >>> from measured.si import Kilo, Meter
            >>> assert Kilo * Meter != 1000 * Meter
            >>> assert 1 * Kilo * Meter == 1000 * Meter
            >>> assert (Kilo * Meter).quantify() == 1000 * Meter
        """
        return self.prefix.quantify() * Unit(
            IdentityPrefix, self.factors, self.dimension
        )

    def __repr__(self) -> str:
        if self.name:
            return f"Unit.named({self.name!r})"

        return (
            "Unit("
            f"prefix={self.prefix!r}, "
            f"factors={self.factors!r}, "
            f"dimension={self.dimension!r}, "
            f"name={self.name!r}, symbol={self.symbol!r}"
            ")"
        )

    def __str__(self) -> str:
        if self.symbol:
            return f"{self.prefix}{self.symbol}"

        return str(self.prefix) + "⋅".join(
            f"{unit.prefix}{unit.symbol}{superscript(exponent)}"
            for unit, exponent in self.factors.items()
        )

    def __format__(self, format_specifier: str) -> str:
        if not format_specifier:
            return str(self)

        if format_specifier == "/":
            numerator, denominator = self.as_ratio()
            if denominator == One:
                return str(self)
            return str(numerator) + "/" + str(denominator)

        raise ValueError(f"Unrecognized format specifier {format_specifier!r}")

    @classmethod
    def _simplify(cls, factors: Mapping["Unit", int]) -> Dict["Unit", int]:
        simplified = {
            unit: exponent
            for unit, exponent in factors.items()
            if unit is not One and exponent != 0
        }
        return simplified or {One: 1}

    def __add__(self, other: "Unit") -> "Unit":
        if self is not other:
            return NotImplemented

        return self

    def __sub__(self, other: "Unit") -> "Unit":
        if self is not other:
            return NotImplemented

        return self

    @staticmethod
    @lru_cache(maxsize=None)
    def _multiply(self: "Unit", other: "Unit") -> "Unit":
        dimension = self.dimension * other.dimension
        prefix = self.prefix * other.prefix

        factors: Dict["Unit", int] = defaultdict(int, self.factors)
        for unit, exponent in other.factors.items():
            factors[unit] += exponent
        factors = self._simplify(factors)

        return Unit(prefix, factors, dimension)

    @overload
    def __mul__(self, other: "Unit") -> "Unit":
        ...  # pragma: no cover

    @overload
    def __mul__(self, other: Numeric) -> "Quantity":
        ...  # pragma: no cover

    def __mul__(self, other: Union["Unit", Numeric]) -> Union["Unit", "Quantity"]:
        if isinstance(other, NUMERIC_CLASSES):
            return Quantity(other, self)

        if not isinstance(other, Unit):
            return NotImplemented

        return Unit._multiply(self, other)

    __rmul__ = __mul__

    @staticmethod
    @lru_cache(maxsize=None)
    def _divide(self: "Unit", other: "Unit") -> "Unit":
        dimension = self.dimension / other.dimension
        prefix = self.prefix / other.prefix

        factors: Dict["Unit", int] = defaultdict(int, self.factors)
        for unit, exponent in other.factors.items():
            factors[unit] -= exponent
        factors = self._simplify(factors)

        return Unit(prefix, factors, dimension)

    def __truediv__(self, other: "Unit") -> "Unit":
        if not isinstance(other, Unit):
            return NotImplemented

        return Unit._divide(self, other)

    def __pow__(self, power: int) -> "Unit":
        if not isinstance(power, int):
            return NotImplemented

        dimension = self.dimension**power
        prefix = self.prefix**power

        factors = self._simplify(
            {unit: exponent * power for unit, exponent in self.factors.items()}
        )
        return Unit(prefix, factors, dimension)

    def root(self, degree: int) -> "Unit":
        """Returns the nth root of this Unit"""
        if not isinstance(degree, int):
            raise TypeError(f"degree should be an integer, not {type(degree)}")

        if degree == 0:
            return One

        dimension = self.dimension.root(degree)
        prefix = self.prefix.root(degree)

        if any(
            exponent // degree != exponent / degree
            for unit, exponent in self.factors.items()
            if exponent > 0 and unit is not One
        ):
            raise FractionalDimensionError(degree, self)

        factors = self._simplify(
            {unit: int(exponent // degree) for unit, exponent in self.factors.items()}
        )
        return Unit(prefix, factors, dimension)

    def as_ratio(self) -> Tuple["Unit", "Unit"]:
        """Returns this unit, split into a numerator and denominator"""
        numerator, denominator = self.dimension.as_ratio()
        return (
            Unit(
                self.prefix,
                {u: e for u, e in self.factors.items() if e >= 0} or {One: 1},
                numerator,
            ),
            Unit(
                IdentityPrefix,
                {u: -e for u, e in self.factors.items() if e < 0} or {One: 1},
                denominator,
            ),
        )


class ConversionNotFound(ValueError):
    pass


@total_ordering
class Quantity:
    """Quantity represents a quantity of some Unit

    Attributes:

        magnitude (int | float): the quantity

        unit (Unit): the [`Unit`][measured.Unit]

    Examples:

        All of the arithmetic operations are supported between Quantities.
        Multiplication and division with [`Unit`][measured.Unit], `int`, and `float`
        are also supported.

        >>> from measured.si import Meter, Second
        >>> assert 2 * Meter + 3 * Meter == 5 * Meter
        >>> assert 3 * Meter - 1 * Meter == 2 * Meter
        >>> assert (2 * Meter) / (1 * Second) == 2 * Meter / Second
        >>> assert 10 * Meter / 2 == 5 * Meter
        >>> assert 10 * Meter * 2 == 20 * Meter
        >>> assert (10 * Meter)**2 == 100 * Meter**2

        Keep an eye on Python's operator precedence, which may lead to surprising
        results.  Consider the following:

        >>> assert (2 * Meter / 1 * Second) != (2 * Meter) / (1 * Second)

        In the above example, the order of operations on the left is:

        * `2 * Meter` → `Quantity(2, Meter)`
        * `Quantity(2, Meter) / 1` → `Quantity(2, Meter)`
        * `Quantity(2, Meter) * Second` → `Quantity(2, Meter * Second)`

        But on the right, the order is:

        * `2 * Meter` → `Quantity(2, Meter)`
        * `1 * Second` → `Quantity(1, Second)`
        * `Quantity(2, Meter) / Quantity(1, Second)` → `Quantity(2, Meter / Second)`


    """

    def __init__(self, magnitude: Numeric, unit: Unit):
        self.magnitude = magnitude
        self.unit = unit

    def in_base_units(self) -> "Quantity":
        """Reduces this Quantity into a new Quantity expressed only in base units
        without any Prefixes"""
        return self.magnitude * self.unit.quantify()

    def in_unit(self, other: Unit) -> "Quantity":
        """Convert this Quantity into another unit"""
        return conversions.convert(self, other)

    # JSON support

    def __json__(self) -> Dict[str, Any]:
        return {
            "magnitude": self.magnitude,
            "unit": self.unit,
        }

    @classmethod
    def __from_json__(cls, json_object: Dict[str, Any]) -> "Quantity":
        return Quantity(json_object["magnitude"], json_object["unit"])

    # Pydantic support

    @classmethod
    def __get_validators__(
        cls,
    ) -> Generator[Callable[["Quantity"], "Quantity"], None, None]:
        yield cls.validate

    @classmethod
    def validate(cls, value: Union[str, "Quantity"]) -> "Quantity":
        if isinstance(value, Quantity):
            return value

        raise ValueError(f"No conversion from {value!r} to Quantity")

    def __repr__(self) -> str:
        return f"Quantity(magnitude={self.magnitude!r}, unit={self.unit!r})"

    def __str__(self) -> str:
        return f"{self.magnitude} {self.unit}"

    def __format__(self, format_specifier: str) -> str:
        magnitude_format, _, unit_format = format_specifier.partition(":")
        return (
            self.magnitude.__format__(magnitude_format)
            + " "
            + self.unit.__format__(unit_format)
        )

    def __add__(self, other: "Quantity") -> "Quantity":
        if isinstance(other, Quantity):
            other = other.in_unit(self.unit)
            return Quantity(self.magnitude + other.magnitude, self.unit)

        return NotImplemented

    def __sub__(self, other: "Quantity") -> "Quantity":
        if isinstance(other, Quantity):
            other = other.in_unit(self.unit)
            return Quantity(self.magnitude - other.magnitude, self.unit)

        return NotImplemented

    def __mul__(self, other: Union["Quantity", "Unit", "Numeric"]) -> "Quantity":
        if isinstance(other, Unit):
            return Quantity(self.magnitude, self.unit * other)

        if isinstance(other, Quantity):
            return Quantity(self.magnitude * other.magnitude, self.unit * other.unit)

        if isinstance(other, NUMERIC_CLASSES):
            return Quantity(self.magnitude * other, self.unit)

        return NotImplemented

    __rmul__ = __mul__

    def __truediv__(self, other: Union["Quantity", "Unit", "Numeric"]) -> "Quantity":
        if isinstance(other, Unit):
            return Quantity(self.magnitude, self.unit / other)

        if isinstance(other, Quantity):
            return Quantity(self.magnitude / other.magnitude, self.unit / other.unit)

        if isinstance(other, NUMERIC_CLASSES):
            return Quantity(self.magnitude / other, self.unit)

        return NotImplemented

    def __pow__(self, power: int) -> "Quantity":
        return Quantity(self.magnitude**power, self.unit**power)

    def root(self, degree: int) -> "Quantity":
        """Returns the nth root of this Quantity"""
        if degree == 0:
            return 1 * One
        return Quantity(self.magnitude ** (1 / degree), self.unit.root(degree))

    def __neg__(self) -> "Quantity":
        return Quantity(-self.magnitude, self.unit)

    def __pos__(self) -> "Quantity":
        return Quantity(+self.magnitude, self.unit)

    def __abs__(self) -> "Quantity":
        return Quantity(abs(self.magnitude), self.unit)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Quantity):
            return NotImplemented

        if self.unit.dimension != other.unit.dimension:
            return NotImplemented

        this = self.in_base_units()
        other = other.in_base_units()

        if this.unit == other.unit:
            return this.magnitude == other.magnitude

        try:
            return this.in_unit(other.unit) == other
        except ConversionNotFound:
            return NotImplemented

    def __lt__(self, other: Any) -> bool:
        if not isinstance(other, Quantity):
            return NotImplemented

        if self.unit.dimension != other.unit.dimension:
            return NotImplemented

        this = self.in_base_units()
        other = other.in_base_units()

        if this.unit == other.unit:
            return this.magnitude < other.magnitude

        try:
            return this.in_unit(other.unit) < other
        except ConversionNotFound:
            return NotImplemented

    def approximates(self, other: "Quantity", within: Numeric = 1e-6) -> bool:
        """Indicates whether this Quantity and another Quantity are close enough to
        each other to be considered equal.

        Parameters:
            other (Quantity): the other quantity to compare this quantity to
            within (int | float): the tolerance

        Examples:

            >>> from measured.si import Meter
            >>> assert (0.001 * Meter).approximates(0.002 * Meter, within=0.01)
        """
        if self == other:
            return True

        if self.unit.dimension != other.unit.dimension:
            return False

        this = self.in_base_units()
        other = other.in_base_units()

        if this.unit != other.unit:
            try:
                this = this.in_unit(other.unit)
            except ConversionNotFound:
                return False

        difference = this - other
        tolerance = Quantity(within, this.unit)

        return abs(difference) <= tolerance

    def assert_approximates(self, other: "Quantity", within: Numeric = 1e-6) -> None:
        """Asserts whether this Quantity and another Quantity are close enough to
        each other to be considered equal, with a helpful assertion message

        Parameters:
            other (Quantity): the other quantity to compare this quantity to
            within (int | float): the tolerance

        Examples:

            >>> from measured.si import Meter
            >>> (0.001 * Meter).assert_approximates(0.002 * Meter, within=0.01)
        """
        left = self.in_base_units()
        right = other.in_base_units()

        message = " or ".join(
            [
                f"{left} !~ {right.in_unit(left.unit)}",
                f"{right} !~ {left.in_unit(right.unit)}",
            ]
        )

        assert self.approximates(other, within=within), message


Ratio = Numeric
Offset = Numeric


class ConversionTable:
    _ratios: Dict[Unit, Dict[Unit, Ratio]]
    _offsets: Dict[Unit, Dict[Unit, Offset]]

    def __init__(self) -> None:
        self._ratios = defaultdict(dict)
        self._offsets = defaultdict(dict)

    def equate(self, a: Quantity, b: Quantity) -> None:
        """Defines a conversion between one Unit and another, expressed as a ratio
        between the two."""

        if a.unit == b.unit and a.unit is not One:
            raise ValueError("No need to define conversions for a unit and itself")

        a = a.in_base_units()
        b = b.in_base_units()

        self._ratios[a.unit][b.unit] = b.magnitude / a.magnitude
        self._ratios[b.unit][a.unit] = a.magnitude / b.magnitude

    def translate(self, scale: Unit, zero: Quantity) -> None:
        """Defines a unit as a scale starting from the given zero point in another
        unit"""
        if scale == zero.unit:
            raise ValueError("No need to define conversions for a unit and itself")

        degree = zero.unit
        offset = zero.magnitude

        self._ratios[degree][scale] = 1
        self._ratios[scale][degree] = 1

        self._offsets[degree][scale] = -offset
        self._offsets[scale][degree] = +offset

    def convert(self, quantity: Quantity, other_unit: Unit) -> Quantity:
        """Converts the given quantity into another unit, if possible"""
        if quantity.unit.dimension != other_unit.dimension:
            raise ConversionNotFound(
                "No conversion from "
                f"{quantity.unit} ({quantity.unit.dimension}) to "
                f"{other_unit} ({other_unit.dimension})"
            )

        this = quantity.in_base_units()
        other = (1 * other_unit).in_base_units()

        this = self._collapse_by_dimension(this)
        other = self._collapse_by_dimension(other)

        this_numerator, this_denominator = this.unit.as_ratio()
        other_numerator, other_denominator = other.unit.as_ratio()

        numerator_path = self._find(this_numerator, other_numerator)
        if not numerator_path:
            raise ConversionNotFound(
                f"No conversion from {this_numerator} to {other_numerator}"
            )

        denominator_path = self._find(this_denominator, other_denominator)
        if not denominator_path:
            raise ConversionNotFound(
                f"No conversion from {this_denominator} to {other_denominator}"
            )

        numerator = this.magnitude
        for scale, offset, _ in numerator_path:
            numerator *= scale
            numerator += offset

        denominator = other.magnitude
        for scale, offset, _ in denominator_path:
            denominator *= scale
            denominator += offset

        return Quantity(numerator / denominator, other_unit)

    def _find(
        self,
        start: Unit,
        end: Unit,
    ) -> Optional[Iterable[Tuple[Ratio, Offset, Unit]]]:
        start_terms = self._terms_by_dimension(start)
        end_terms = self._terms_by_dimension(end)

        assert (
            start_terms.keys() == end_terms.keys()
        ), f"{start_terms.keys()} != {end_terms.keys()}"

        path: List[Tuple[Ratio, Offset, Unit]] = []
        for dimension in start_terms:
            for s, e in zip(start_terms[dimension], end_terms[dimension]):
                this_path = self._find_path(s, e)
                if not this_path:
                    return None
                path += this_path
        return path

    @classmethod
    def _terms_by_dimension(cls, unit: Unit) -> Dict[Dimension, List[Unit]]:
        terms = defaultdict(list)
        for factor, exponent in unit.factors.items():
            factor = factor**exponent
            terms[factor.dimension].append(factor)
        return terms

    def _collapse_by_dimension(self, quantity: Quantity) -> Quantity:
        """Return a new quantity with at most a single unit in each dimension, by
        converting individual terms"""
        magnitude = quantity.magnitude
        by_dimension: Dict[Dimension, Tuple[Unit, int]] = {}
        for unit, exponent in quantity.unit.factors.items():
            dimension = unit.dimension
            quantified = unit.quantify()

            if dimension not in by_dimension:
                magnitude *= quantified.magnitude**exponent
                by_dimension[dimension] = (quantified.unit, exponent)
                continue

            current_unit, current_exponent = by_dimension[dimension]

            path = self._find_path(quantified.unit, current_unit)
            if not path:
                raise ConversionNotFound(
                    f"No conversion between {dimension} units {quantified.unit} "
                    f"and {current_unit}"
                )

            for scale, offset, _ in path:
                magnitude *= scale**exponent
                magnitude += offset

            by_dimension[dimension] = (current_unit, current_exponent + exponent)

        factors = {
            unit: exponent for unit, exponent in by_dimension.values() if exponent != 0
        } or {One: 1}

        final_dimension = Number
        for dimension in by_dimension.keys():
            final_dimension *= dimension

        final = Quantity(magnitude, Unit(IdentityPrefix, factors, final_dimension))
        return final

    def _find_path(
        self,
        start: Unit,
        end: Unit,
        visited: Optional[Set[Unit]] = None,
    ) -> Optional[List[Tuple[Ratio, Offset, Unit]]]:

        if start is end:
            return [(1, 0, end)]

        if visited is None:
            visited = {start}
        elif start in visited:
            return None
        else:
            visited.add(start)

        exponent, start, end = self._reduce_dimension(start, end)

        if sum(start.factors.values()) > sum(end.factors.values()):
            # This is a conversion like m² -> acre, where the end dimension is defined
            # directly in the higher exponent and there isn't a lower-power unit (e.g.
            # there's no unit that represents the square root of an acre that we can
            # compare the meter to); in this case, perform the search in reverse and it
            # should be able to find available conversions
            backtracked = self._backtrack(self._find_path(end, start), exponent, end)
            return backtracked

        best_path = None

        for intermediate, scale in self._ratios[start].items():
            offset = self._offsets[start].get(intermediate, 0)
            if intermediate == end:
                return [(scale**exponent, offset**exponent, end**exponent)]

            path = self._find_path(intermediate, end, visited=visited)
            if not path:
                continue

            path = [(scale, offset, intermediate)] + list(path)
            path = [
                (scale**exponent, offset**exponent, unit**exponent)
                for scale, offset, unit in path
            ]
            if not best_path or len(path) < len(best_path):
                best_path = path

        if best_path:
            return best_path

        return None

    @classmethod
    def _reduce_dimension(cls, start: Unit, end: Unit) -> Tuple[int, Unit, Unit]:
        """Reduce the dimension of the given units to their lowest common exponents"""
        assert (
            start.dimension == end.dimension
        ), f"{start} and {end} measure different dimensions"

        if start.dimension == Number:
            return 1, start, end

        exponent = gcd(*start.dimension.exponents)

        try:
            start_root = start.root(exponent)
            end_root = end.root(exponent)
        except FractionalDimensionError:
            return 1, start, end

        return exponent, start_root, end_root

    @classmethod
    def _backtrack(
        cls,
        path: Optional[Iterable[Tuple[Ratio, Offset, Unit]]],
        exponent: int,
        end: Unit,
    ) -> Optional[List[Tuple[Ratio, Offset, Unit]]]:
        """Given a path to convert a start unit to an end unit, produce the reverse
        path, which would convert the end unit to the start unit"""
        if path is None:
            return None

        path = list(reversed(list(path)))

        units = [u for _, _, u in path[1:]] + [end]
        scales_and_offsets = [(s, o) for s, o, _ in path]

        backtracked = [
            (1 / (scale**exponent), -(offset**exponent), unit**exponent)
            for (scale, offset), unit in zip(scales_and_offsets, units)
        ]
        return backtracked


conversions = ConversionTable()


# https://en.wikipedia.org/wiki/Dimensional_analysis#Definition

# Fundamental physical dimensions

Number = Dimension.define(name="number", symbol="1")
Length = Dimension.define(name="length", symbol="L")
Time = Dimension.define(name="time", symbol="T")
Mass = Dimension.define(name="mass", symbol="M")
Current = Dimension.define(name="current", symbol="I")
Temperature = Dimension.define(name="temperature", symbol="Θ")
AmountOfSubstance = Dimension.define(name="amount of substance", symbol="N")
LuminousIntensity = Dimension.define(name="luminous intensity", symbol="J")
Information = Dimension.define(name="information", symbol="B")  # TODO


# Derived dimensions

Area = Dimension.derive(Length * Length, name="area")
Volume = Dimension.derive(Area * Length, name="volume")

PlaneAngle = Length / Length
SolidAngle = Area / Area

# https://en.wikipedia.org/wiki/Fourth,_fifth,_and_sixth_derivatives_of_position

Speed = Dimension.derive(Length / Time, name="speed")
Acceleration = Dimension.derive(Length / Time**2, name="acceleration")
Jerk = Dimension.derive(Length / Time**3, name="jerk")
Snap = Dimension.derive(Length / Time**4, name="snap")
Crackle = Dimension.derive(Length / Time**5, name="crackle")
Pop = Dimension.derive(Length / Time**6, name="pop")

Frequency = Dimension.derive(Number / Time, name="frequency")

# https://en.wikipedia.org/wiki/Newton%27s_laws_of_motion#Second

Force = Dimension.derive(Mass * Acceleration, name="force")
Energy = Dimension.derive(Length * Force, name="energy")
Power = Dimension.derive(Energy / Time, name="power")

Charge = Dimension.derive(Time * Current, name="charge")
Potential = Dimension.derive(Energy / Charge, name="potential")
Capacitance = Dimension.derive(Charge / Potential, name="capacitance")
Resistance = Dimension.derive(Potential / Current, name="resistance")
Conductance = Dimension.derive(Current / Potential, name="conductance")
Inductance = Dimension.derive(Potential * Time / Current, name="inductance")

MagneticFlux = Dimension.derive(Potential * Time, name="magnetic flux")
MagneticBField = Dimension.derive(Potential * Time / Area, name="magnetic B-field")

LuminousFlux = LuminousIntensity * SolidAngle
Illuminance = Dimension.derive(LuminousIntensity / Area, name="illuminance")

RadioactiveDose = Dimension.derive(Power / Mass, name="radioactivedose")

Catalysis = Dimension.derive(AmountOfSubstance / Time, name="catalysis")


# Fundamental prefixes

IdentityPrefix = Prefix(0, 0)


# Fundamental units

One = Number.unit(name="one", symbol="1")
One.equals(1 * One)



from measured import (
    AmountOfSubstance,
    Area,
    Current,
    Frequency,
    Length,
    LuminousIntensity,
    Mass,
    One,
    Temperature,
    Time,
    Unit,
    Volume,
)
from measured.si import Ampere, Candela, Gram, Hertz, Kelvin, Meter, Mole, Second


def test_meter() -> None:
    assert isinstance(Meter, Unit)
    assert Meter.dimension is Length
    assert Meter.name == "meter"
    assert Meter.symbol == "m"


def test_square_meter() -> None:
    square_meter = Meter**2
    assert square_meter.dimension is Area
    assert square_meter.factors == {Meter: 2}
    assert square_meter is Meter * Meter


def test_cubic_meter() -> None:
    cubic_meter = Meter**3
    assert cubic_meter.dimension is Volume
    assert cubic_meter.factors == {Meter: 3}
    assert cubic_meter is Meter * Meter * Meter


def test_second() -> None:
    assert isinstance(Second, Unit)
    assert Second.dimension is Time
    assert Second.name == "second"
    assert Second.symbol == "s"


def test_hertz() -> None:
    assert Hertz.dimension is Frequency
    assert Hertz.factors == {Second: -1}
    assert Hertz.name == "hertz"
    assert Hertz.symbol == "Hz"
    assert Hertz is One / Second
    assert Hertz is Second**-1


def test_gram() -> None:
    assert isinstance(Gram, Unit)
    assert Gram.dimension is Mass
    assert Gram.name == "gram"
    assert Gram.symbol == "g"


def test_ampere() -> None:
    assert isinstance(Ampere, Unit)
    assert Ampere.dimension is Current
    assert Ampere.name == "ampere"
    assert Ampere.symbol == "A"


def test_kelvin() -> None:
    assert isinstance(Kelvin, Unit)
    assert Kelvin.dimension is Temperature
    assert Kelvin.name == "kelvin"
    assert Kelvin.symbol == "K"


def test_mole() -> None:
    assert isinstance(Mole, Unit)
    assert Mole.dimension is AmountOfSubstance
    assert Mole.name == "mole"
    assert Mole.symbol == "mol"


def test_candela() -> None:
    assert isinstance(Candela, Unit)
    assert Candela.dimension is LuminousIntensity
    assert Candela.name == "candela"
    assert Candela.symbol == "cd"



from math import pi as π

from measured.astronomical import (
    AstronomicalUnit,
    EarthMass,
    JulianYear,
    JupiterMass,
    LightYear,
    Parsec,
    SolarMass,
)
from measured.fundamental import GravitationalConstant
from measured.si import Meter


def test_parsec() -> None:
    # https://en.wikipedia.org/wiki/Parsec#Calculating_the_value_of_a_parsec
    (π * Parsec).assert_approximates(180 * 60 * 60 * AstronomicalUnit)
    (1 * Parsec).assert_approximates(3.0856775814913673e16 * Meter, within=1e3)


def test_light_year() -> None:
    # https://en.wikipedia.org/wiki/Light-year
    (1 * LightYear).assert_approximates(9460730472580800 * Meter)
    (1 * LightYear).assert_approximates(63241.077 * AstronomicalUnit, within=1e3)
    (1 * LightYear).assert_approximates(0.306601 * Parsec)


def test_deriving_solar_mass() -> None:
    # https://en.wikipedia.org/wiki/Solar_mass#Calculation
    calculated = (4 * π**2 * AstronomicalUnit**3) / (
        GravitationalConstant * JulianYear**2
    )
    calculated.assert_approximates(1 * SolarMass, within=2.74e27)


def test_solar_system_masses() -> None:
    assert 1 * EarthMass < 1 * JupiterMass < 1 * SolarMass
    (1 * SolarMass).assert_approximates(1047.348644 * JupiterMass, within=1e0)
    (1 * SolarMass).assert_approximates(332946.0487 * EarthMass, within=1e1)
    (1 * JupiterMass).assert_approximates(317.82838 * EarthMass, within=1e-3)



"""
Defines units and constants of use in the field of astronomy, which requires measuring
extremely large distances, masses, and periods of time.

> The [astronomical system of units][1] is a tridimensional system, in that it defines
> units of length, mass and time.  The associated astronomical constants also fix the
> different frames of reference that are needed to report observations.

[1]: https://en.wikipedia.org/wiki/Astronomical_system_of_units

Attributes: Units of length

    AstronomicalUnit (Unit): A unit representing the average distance between the Earth
        and the Sun; it is now defined precisely as 149,597,870,700 meters

    LightYear (Unit): the distance light travels through a vacuum in one `JulianYear`

    Parsec (Unit): A measure of large astronomical distances defined in terms of
        [trigonometry and parallax][1]

        [1]: https://en.wikipedia.org/wiki/Parsec

Attributes: Units of time

    Day (Unit): Uses the same definition of `Day` as the SI system

    JulianYear (Unit): defined as 365.25 days

Attributes: Units of mass

    SolarMass (Unit): The mass of the Sun

    EarthMass (Unit): The mass of the Earth

    JupiterMass (Unit): The mass of Jupiter
"""

from math import pi as π

from measured import Length, Mass, Time, Unit, si
from measured.fundamental import SpeedOfLight
from measured.si import Kilogram, Meter

# Measures of Time

Day = Unit.derive(si.Day, name="day", symbol="D")

JulianYear = Time.unit("Julian year", "a")
JulianYear.equals(365.25 * Day)

# Measures of Length

AstronomicalUnit = Length.unit("astronomical unit", "AU")
AstronomicalUnit.equals(149597870700 * Meter)

# https://en.wikipedia.org/wiki/Light-year
LightYear = Length.unit("light-year", "ly")
LightYear.equals((SpeedOfLight * JulianYear).in_unit(Meter))

# https://en.wikipedia.org/wiki/Parsec#Calculating_the_value_of_a_parsec
Parsec = Length.unit("parsec", "pc")
Parsec.equals(96939420213600600 / π * Meter)

# Measures of Mass

# https://en.wikipedia.org/wiki/Solar_mass
SolarMass = Mass.unit("solar mass", "M☉")
SolarMass.equals(1.98847e30 * Kilogram)

EarthMass = Mass.unit("earth mass", "M-earth")
EarthMass.equals(5.9722e24 * Kilogram)

JupiterMass = Mass.unit("jupiter mass", "M-jup")
JupiterMass.equals(1.89813e27 * Kilogram)



import pytest

from measured import Quantity
from measured.si import Kilo, Meter
from measured.us import (
    Acre,
    Chain,
    Foot,
    League,
    Rod,
    Section,
    StatuteMile,
    SurveyFoot,
    SurveyTownship,
)


@pytest.mark.parametrize(
    "left, right",
    [
        (1 * Chain**2, 4356 * SurveyFoot**2),
        (1 * Chain**2, 16 * Rod**2),
        (1 * Acre, 43560 * Foot**2),
        (1 * Acre, 10 * Chain**2),
        (1 * Section, 640 * Acre),
        (1 * Section, 1 * StatuteMile**2),
        (1 * SurveyTownship, 36 * Section),
        (1 * SurveyTownship, 4 * League**2),
    ],
)
def test_areas_equal(left: Quantity, right: Quantity) -> None:
    assert left == right, f"{left} != {right.in_unit(left.unit)}"
    assert right == left, f"{right} != {left.in_unit(right.unit)}"


@pytest.mark.parametrize(
    "left, right, within",
    [
        (1 * Chain**2, 404.6873 * Meter**2, 1e-4),
        (1 * Acre, 4046.86 * Meter**2, 1e-2),
        (1 * Section, 2.589998 * (Kilo * Meter) ** 2, 1e0),
        (1 * SurveyTownship, 93.23993 * (Kilo * Meter) ** 2, 1e2),
    ],
)
def test_areas_approximate(left: Quantity, right: Quantity, within: float) -> None:
    left.assert_approximates(right, within)
    right.assert_approximates(left, within)



import pytest

from measured import (
    Area,
    ConversionNotFound,
    Length,
    Mass,
    Number,
    Numeric,
    One,
    Unit,
    conversions,
)
from measured.si import Meter, Second
from measured.us import Acre, Foot, Inch


def test_disallow_self_conversion() -> None:
    with pytest.raises(ValueError, match="unit and itself"):
        Meter.equals(1 * Meter)

    with pytest.raises(ValueError, match="unit and itself"):
        conversions.equate(1 * Meter, 1 * Meter)

    with pytest.raises(ValueError, match="unit and itself"):
        conversions.equate(1 * Meter, 2 * Meter)


def test_disallow_self_zeroing() -> None:
    with pytest.raises(ValueError, match="unit and itself"):
        Meter.zero(1 * Meter)

    with pytest.raises(ValueError, match="unit and itself"):
        conversions.translate(Meter, 1 * Meter)


def test_approximating_requires_units_with_conversion() -> None:
    Bogus = Length.unit(name="bogus", symbol="bog")

    one = 1 * Meter
    other = 1 * Bogus

    assert not one.approximates(other, within=1e100)


def test_unit_conversion_must_be_in_same_dimension() -> None:
    with pytest.raises(ConversionNotFound):
        (1 * Meter).in_unit(Second)


def test_unit_always_converts_to_itself() -> None:
    converted = (10 * Meter).in_unit(Meter)
    assert converted.magnitude == 10
    assert converted.unit == Meter


def test_conversion_with_exponents() -> None:
    assert (1 * Foot) ** 2 == (12 * Inch) ** 2
    assert 1 * Foot**2 == 144 * Inch**2

    assert (1 * Foot) ** 3 == (12 * Inch) ** 3
    assert 1 * Foot**3 == 1728 * Inch**3


@pytest.mark.parametrize(
    "start, end, magnitude",
    [
        (Foot, Inch, 12.0),
        (Foot**2, Meter**2, 0.09290304),
        (Meter**2, Foot**2, 10.76391041670972),
        (Acre, Foot**2, 43560.0),
        (Foot**2, Inch**2, 144.0),
        (Acre, Inch**2, 43560.0 * 144.0),
        (Inch**2, Foot**2, 0.006944444444444444),
        (Foot**2, Acre, 0.00002295684113865932),
        (Inch**2, Acre, 0.006944444444444444 * 0.00002295684113865932),
    ],
)
def test_conversion_can_navigate_exponents(
    start: Unit, end: Unit, magnitude: Numeric
) -> None:
    converted = conversions.convert(1 * start, end)
    assert converted == magnitude * end
    assert converted.unit == end
    assert converted.magnitude == magnitude


def test_backtracking_conversions_with_no_path() -> None:
    # Set up some arbitrary units that don't convert to one another
    Hiya = Length.unit("hello", "hello")
    Mundo = Area.unit("world", "world")

    with pytest.raises(ConversionNotFound):
        conversions.convert(1 * Hiya**2, Mundo)

    with pytest.raises(ConversionNotFound):
        conversions.convert(1 * Mundo, Hiya**2)


def test_failing_to_convert_numerator() -> None:
    Gud = Length.unit("goodbye", "goodbye")
    Mun = Length.unit("moon", "moon")

    with pytest.raises(ConversionNotFound):
        (1 * Gud).in_unit(Mun)


def test_failing_to_convert_denominator() -> None:
    Flib = Mass.unit("flibbidy", "flibbidy")
    Flob = Mass.unit("flobbidy", "flobbidy")

    assert (Meter / Flib).dimension == (Meter / Flob).dimension

    with pytest.raises(ConversionNotFound):
        (1 * Meter / Flib).in_unit(Meter / Flob)


def test_failing_to_collapse_dimensions() -> None:
    Jib = Mass.unit("jibbity", "jibbity")
    Job = Mass.unit("jobbity", "jobbity")

    assert (Jib / Job).dimension == Number
    assert (Job / Jib).dimension == Number

    with pytest.raises(ConversionNotFound):
        ((1 * Jib) / (1 * Job)).in_unit(One)



import pytest

from measured import Information, One, Quantity
from measured.iec import Bit, Byte, Gibi, Kibi, Mebi
from measured.si import Giga, Kilo, Mega, Second


def test_bits_measure_information() -> None:
    assert Bit.dimension == Information
    assert Bit.name == "bit"
    assert Bit.symbol == "b"


def test_bytes_measure_information() -> None:
    assert Byte.dimension == Information
    assert Byte.name == "byte"
    assert Byte.symbol == "B"


def test_bytes_are_8_bits() -> None:
    assert 8 * Bit == 1 * Byte
    assert 1 * (Byte / Bit) == 8 * One
    assert 1 * Byte / Bit == 8 * One


def test_iec_prefixes() -> None:
    assert 1024 * (Kibi * Byte) == 1 * (Mebi * Byte)


def test_data_rates() -> None:
    assert 1024 * (Mebi * Byte) / Second == 1 * (Gibi * Byte / Second)


@pytest.mark.parametrize(
    "iec, si",
    [
        (1 * (Kilo * Bit), 0.9765625 * (Kibi * Bit)),
        (1 * (Mega * Bit), 0.95367431640625 * (Mebi * Bit)),
        (1 * (Giga * Bit), 0.9313225746154785 * (Gibi * Bit)),
    ],
)
def test_bit_iec_si_compatibility(iec: Quantity, si: Quantity) -> None:
    assert iec == si


@pytest.mark.parametrize(
    "iec, si",
    [
        (1 * (Kilo * Byte), 0.9765625 * (Kibi * Byte)),
        (1 * (Mega * Byte), 0.95367431640625 * (Mebi * Byte)),
        (1 * (Giga * Byte), 0.9313225746154785 * (Gibi * Byte)),
        (1 * (Kilo * Byte), 8 * 0.9765625 * (Kibi * Bit)),
        (1 * (Mega * Byte), 8 * 0.95367431640625 * (Mebi * Bit)),
        (1 * (Giga * Byte), 8 * 0.9313225746154785 * (Gibi * Bit)),
        (8 * (Kilo * Bit), 0.9765625 * (Kibi * Byte)),
        (8 * (Mega * Bit), 0.95367431640625 * (Mebi * Byte)),
        (8 * (Giga * Bit), 0.9313225746154785 * (Gibi * Byte)),
    ],
)
def test_byte_iec_si_compatibility(iec: Quantity, si: Quantity) -> None:
    difference = iec.in_base_units() - si.in_base_units()
    assert iec.approximates(si, within=1e-5), f"{iec} and {si} differ by {difference}"



"""
Defines the units of `Information` and the binary [prefixes][2] as defined by the
International Electrotechnical Commission.

[1]: https://en.wikipedia.org/wiki/Byte

[2]: https://en.wikipedia.org/wiki/Binary_prefix#Adoption_by_IEC,_NIST_and_ISO

Attributes: Base Units

    Bit (Unit):

Attributes: Derived Units

    Byte (Unit): 8 bits

Attributes: Prefixes (base 2)

    Kibi (Prefix): 2¹⁰, symbol `Ki`
    Mebi (Prefix): 2²⁰, symbol `Mi`
    Gibi (Prefix): 2³⁰, symbol `Gi`
    Tebi (Prefix): 2⁴⁰, symbol `Ti`
    Pebi (Prefix): 2⁵⁰, symbol `Pi`
    Exbi (Prefix): 2⁶⁰, symbol `Ei`
    Zebi (Prefix): 2⁷⁰, symbol `Zi`
    Yobi (Prefix): 2⁸⁰, symbol `Yi`

"""

# https://en.wikipedia.org/wiki/Binary_prefix#Adoption_by_IEC,_NIST_and_ISO

from . import Information, Prefix, Unit

Kibi = Prefix(2, 10, name="kibi", symbol="Ki")
Mebi = Prefix(2, 20, name="mebi", symbol="Mi")
Gibi = Prefix(2, 30, name="gibi", symbol="Gi")
Tebi = Prefix(2, 40, name="tebi", symbol="Ti")
Pebi = Prefix(2, 50, name="pebi", symbol="Pi")
Exbi = Prefix(2, 60, name="exbi", symbol="Ei")
Zebi = Prefix(2, 70, name="zebi", symbol="Zi")
Yobi = Prefix(2, 80, name="yobi", symbol="Yi")

Bit = Information.unit(name="bit", symbol="b")
Byte = Unit.derive(Bit.scale(Prefix(2, 3)), name="byte", symbol="B")



from fractions import Fraction

import pytest

from measured import (
    AmountOfSubstance,
    Area,
    Current,
    Dimension,
    Frequency,
    Length,
    LuminousIntensity,
    Mass,
    Number,
    PlaneAngle,
    SolidAngle,
    Speed,
    Temperature,
    Time,
    Volume,
)


def pytest_generate_tests(metafunc: pytest.Metafunc) -> None:
    fundamental = Dimension.fundamental()
    ids = [d.name for d in fundamental]

    for exemplar in ["a", "b", "c"]:
        if exemplar in metafunc.fixturenames:
            metafunc.parametrize(exemplar, fundamental, ids=ids)

    if "dimension" in metafunc.fixturenames:
        metafunc.parametrize("dimension", fundamental, ids=ids)


@pytest.fixture
def identity() -> Dimension:
    return Number


def test_homogenous_under_addition(a: Dimension, b: Dimension) -> None:
    # https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity
    #
    # Only commensurable quantities (physical quantities having the same dimension) may
    # be compared, equated, added, or subtracted.
    if a is b:
        assert a + b == a
    else:
        with pytest.raises(TypeError):
            a + b


def test_homogenous_under_subtraction(a: Dimension, b: Dimension) -> None:
    # https://en.wikipedia.org/wiki/Dimensional_analysis#Dimensional_homogeneity
    #
    # Only commensurable quantities (physical quantities having the same dimension) may
    # be compared, equated, added, or subtracted.
    if a is b:
        assert a - b == a
    else:
        with pytest.raises(TypeError):
            a - b


def test_abelian_associativity(a: Dimension, b: Dimension, c: Dimension) -> None:
    # https://en.wikipedia.org/wiki/Abelian_group
    assert (a * b) * c == a * (b * c)


def test_abelian_identity(identity: Dimension, a: Dimension) -> None:
    assert identity * a == a


def test_abelian_inverse(identity: Dimension, a: Dimension) -> None:
    inverse = a**-1
    assert inverse * a == a * inverse
    assert inverse * a == identity
    assert identity / a == inverse


def test_abelian_commutativity(a: Dimension, b: Dimension) -> None:
    assert a * b == b * a


def test_no_dimensional_exponentation(dimension: Dimension) -> None:
    with pytest.raises(TypeError):
        dimension**dimension  # type: ignore


def test_no_floating_point_exponentation(dimension: Dimension) -> None:
    with pytest.raises(TypeError):
        dimension**0.5  # type: ignore


def test_no_fractional_exponentation(dimension: Dimension) -> None:
    with pytest.raises(TypeError):
        dimension ** Fraction(1, 2)  # type: ignore


def test_only_dimensional_multiplication(dimension: Dimension) -> None:
    with pytest.raises(TypeError):
        5 * dimension  # type: ignore

    with pytest.raises(TypeError):
        dimension * 5  # type: ignore


def test_only_dimensional_division(dimension: Dimension) -> None:
    with pytest.raises(TypeError):
        dimension / 5  # type: ignore

    with pytest.raises(TypeError):
        5 / dimension  # type: ignore


def test_repr(dimension: Dimension) -> None:
    r = repr(dimension)
    assert r.startswith("Dimension(exponents=(0,")
    assert dimension.name and dimension.name in r
    assert dimension.symbol and dimension.symbol in r
    assert r.endswith(")")


def test_repr_roundtrips(dimension: Dimension) -> None:
    assert eval(repr(dimension)) is dimension


def test_number() -> None:
    assert isinstance(Number, Dimension)
    assert Number.name == "number"
    assert Number.symbol == "1"
    assert Number is Number * Number


def test_length() -> None:
    assert isinstance(Length, Dimension)
    assert Length.name == "length"
    assert Length.symbol == "L"
    assert Length is Number * Length


def test_area() -> None:
    assert isinstance(Area, Dimension)
    assert Area is Length * Length


def test_volume() -> None:
    assert isinstance(Area, Dimension)
    assert Volume is Length * Length * Length
    assert Volume is Length * Area
    assert Volume is Area * Length


def test_time() -> None:
    assert isinstance(Time, Dimension)
    assert Time.name == "time"
    assert Time.symbol == "T"
    assert Time is Number * Time


def test_frequency() -> None:
    assert isinstance(Frequency, Dimension)
    assert Frequency is Number / Time


def test_mass() -> None:
    assert isinstance(Mass, Dimension)
    assert Mass.name == "mass"
    assert Mass.symbol == "M"
    assert Mass is Number * Mass


def test_current() -> None:
    assert isinstance(Current, Dimension)
    assert Current.name == "current"
    assert Current.symbol == "I"
    assert Current is Number * Current


def test_temperature() -> None:
    assert isinstance(Temperature, Dimension)
    assert Temperature.name == "temperature"
    assert Temperature.symbol == "Θ"
    assert Temperature is Number * Temperature


def test_amount_of_substance() -> None:
    assert isinstance(AmountOfSubstance, Dimension)
    assert AmountOfSubstance.name == "amount of substance"
    assert AmountOfSubstance.symbol == "N"
    assert AmountOfSubstance is Number * AmountOfSubstance


def test_luminous_intensity() -> None:
    assert isinstance(LuminousIntensity, Dimension)
    assert LuminousIntensity.name == "luminous intensity"
    assert LuminousIntensity.symbol == "J"
    assert LuminousIntensity is Number * LuminousIntensity


def test_plane_angle_is_dimensionless() -> None:
    assert isinstance(PlaneAngle, Dimension)
    assert PlaneAngle == Number


def test_solid_angle_is_dimensionless() -> None:
    assert isinstance(SolidAngle, Dimension)
    assert SolidAngle == Number


def test_roots() -> None:
    assert Number.root(0) == Number
    assert Number.root(100) == Number
    assert Length.root(0) == Number
    assert Length.root(1) == Length
    assert Area.root(2) == Length
    assert Volume.root(3) == Length
    assert (Area / Time**2).root(2) == Speed


def test_only_integer_roots() -> None:
    with pytest.raises(TypeError):
        Area.root(0.5)  # type: ignore


def test_whole_power_roots_only() -> None:
    with pytest.raises(ValueError):
        Volume.root(2)



import cProfile
import functools
import os
from pathlib import Path
from typing import Callable

import typer

from measured import Quantity
from measured.si import Ampere, Meter, Ohm, Volt

app = typer.Typer()
options = {"baseline": False}

profiles = Path(".profiles")


@app.callback()
def main(baseline: bool = False) -> None:
    options["baseline"] = baseline


def profiled(func: Callable[[], None]) -> Callable[[], None]:
    @functools.wraps(func)
    def inner() -> None:
        os.makedirs(profiles, exist_ok=True)
        filename = f"{'baseline-' if options['baseline'] else ''}{func.__name__}.prof"
        cProfile.runctx("func()", {"func": func}, {}, str(profiles / filename))

    return inner


@app.command()
@profiled
def quantity_construction() -> None:
    """Tests for constructing Quantities"""
    for a in range(1, 1000001):
        Quantity(a, Meter)


@app.command()
@profiled
def low_equality() -> None:
    """Tests for equality for low-dimensional quantities"""
    for a in (Quantity(a, Meter) for a in range(1, 1001)):
        for b in (Quantity(b, Meter) for b in range(1, 1001)):
            bool(a == b)


@app.command()
@profiled
def high_equality() -> None:
    """Tests for equality for high-dimensional quantities"""
    for a in (Quantity(a, Ohm) for a in range(1, 1001)):
        for b in (Quantity(b, Ohm) for b in range(1, 1001)):
            bool(a == b)


@app.command()
@profiled
def resistances() -> None:
    """Computes a whole lot of Ohms"""
    for a in (Quantity(a, Ampere) for a in range(1, 1001)):
        for v in (Quantity(v, Volt) for v in range(1, 1001)):
            assert v / a == (v.magnitude / a.magnitude) * Ohm


if __name__ == "__main__":
    app()



"""
Defines both the [fundamental _dimensionless_ constants][1], as well as the
[foundational _dimensioned_ physical constants][2] of the universe.

[1]: https://en.wikipedia.org/wiki/Dimensionless_physical_constant
[2]: https://en.wikipedia.org/wiki/Physical_constant

Attributes: Physical constants

    SpeedOfLight (Quantity): The speed that light travels unimpeded in a vacuum

    GravitationalConstant (Quantity): The "strength" of the gravitational force
"""

from measured.si import Kilogram, Meter, Second

SpeedOfLight = 299792458 * Meter / Second
GravitationalConstant = 6.6743015e-11 * Meter**3 / (Kilogram * Second**2)



from measured.si import Centi, Deci, Kilo, Liter, Meter, Milli

cm = Centi * Meter
km = Kilo * Meter


def test_length_equivalence() -> None:
    assert 1 * Meter == 100 * cm
    assert 1 * Meter == 0.001 * km


def test_length_conversions() -> None:
    assert (1 * Meter).in_unit(cm).unit is cm
    assert (1 * Meter).in_unit(cm).magnitude == 100

    assert (1 * Meter).in_unit(km).unit is km
    assert (1 * Meter).in_unit(km).magnitude == 0.001


def test_area_equivalence() -> None:
    assert 1 * Meter**2 == 10000 * cm**2
    assert 1 * Meter**2 == 0.000001 * km**2


def test_area_conversions() -> None:
    assert (1 * Meter**2).in_unit(cm**2).unit is cm ** 2
    assert (1 * Meter**2).in_unit(cm**2).magnitude == 10000

    assert (1 * Meter**2).in_unit(km**2).unit is km ** 2
    assert (1 * Meter**2).in_unit(km**2).magnitude == 0.000001


def test_volume_equivalence() -> None:
    assert 1 * Liter == 0.001 * Meter**3
    assert 1 * Liter == 1 * (Deci * Meter) ** 3
    assert 1 * Liter == 1000 * (Milli * Liter)


def test_volume_conversions() -> None:
    assert (1 * Liter).in_unit(Meter**3).unit == Meter**3
    assert (1 * Liter).in_unit(Meter**3).magnitude == 0.001

    assert (1 * Liter).in_unit((Deci * Meter) ** 3).unit == (Deci * Meter) ** 3
    assert (1 * Liter).in_unit((Deci * Meter) ** 3).magnitude == 1



import pytest

from measured import (
    Area,
    Dimension,
    Frequency,
    Length,
    Number,
    Prefix,
    Quantity,
    Unit,
    Volume,
)
from measured.formatting import superscript
from measured.si import Hertz, Kilo, Meter, Milli, Ohm, Second


@pytest.mark.parametrize(
    "exponent, superscripted",
    [
        (-1234567890, "⁻¹²³⁴⁵⁶⁷⁸⁹⁰"),
        (-123.456, "⁻¹²³.⁴⁵⁶"),
        (-123, "⁻¹²³"),
        (-23, "⁻²³"),
        (-3, "⁻³"),
        (-1.1, "⁻¹.¹"),
        (-1.0, "⁻¹.⁰"),
        (-1, "⁻¹"),
        (-0.9, "⁻⁰.⁹"),
        (0, "⁰"),
        (1, ""),
        (0.9, "⁰.⁹"),
        (1.0, ""),
        (1.1, "¹.¹"),
        (2, "²"),
        (23, "²³"),
        (345, "³⁴⁵"),
        (123.456, "¹²³.⁴⁵⁶"),
        (1234567890, "¹²³⁴⁵⁶⁷⁸⁹⁰"),
    ],
)
def test_superscripts(exponent: int, superscripted: str) -> None:
    assert superscript(exponent) == superscripted


def test_formatting_prefixes_simplifies() -> None:
    # This is a little surprising, but what's happening here is that
    # 5 meter per kilosecond is getting reduced to 5 millimeter per second
    assert str((5 * Meter) / (Kilo * Second)) == "5 mm⋅s⁻¹"


def test_do_the_best_we_can_with_odd_prefixes() -> None:
    kilo_plus_one = Prefix(10, 4)
    assert str(5 * (kilo_plus_one * Meter)) == "5 10⁴m"


@pytest.mark.parametrize(
    "dimension, string",
    [
        (Number, "1"),
        (Length, "L"),
        (Area, "L²"),
        (Volume, "L³"),
        (Frequency, "T⁻¹"),
    ],
)
def test_strings_of_dimensions(dimension: Dimension, string: str) -> None:
    assert str(dimension) == string


@pytest.mark.parametrize(
    "unit, string",
    [
        (Meter, "m"),
        (Meter**2, "m²"),
        (Meter**3, "m³"),
        (Meter**3 / Second**2, "m³⋅s⁻²"),
        (Hertz, "Hz"),
        (Hertz**2, "s⁻²"),
    ],
)
def test_strings_of_units(unit: Unit, string: str) -> None:
    assert str(unit) == string


@pytest.mark.parametrize(
    "quantity, string",
    [
        (5 * Meter, "5 m"),
        (5.1 * Meter**2, "5.1 m²"),
        (5 * (Kilo * Meter), "5 km"),
        (5.1 * (Kilo * Meter**2) / Second, "5.1 km²⋅s⁻¹"),
    ],
)
def test_strings_of_quantities(quantity: Quantity, string: str) -> None:
    assert str(quantity) == string


@pytest.mark.parametrize(
    "quantity, template, string",
    [
        (5 * Meter, "{quantity}", "5 m"),
        (5.123 * Meter, "{quantity:.2f:/}", "5.12 m"),
        (5.123 * Meter / Second, "{quantity:.2f}", "5.12 m⋅s⁻¹"),
        (5.123 * Meter / Second, "{quantity:.2f:}", "5.12 m⋅s⁻¹"),
        (5.123 * Meter / Second, "{quantity:.2f:/}", "5.12 m/s"),
        (5 * Ohm, "{quantity}", "5 Ω"),
        (5 * Ohm, "{quantity::/}", "5 m²⋅kg/s³⋅A²"),
        (5 * Milli * Ohm, "{quantity::/}", "0.005 m²⋅kg/s³⋅A²"),
    ],
)
def test_format_specifiers(quantity: Quantity, template: str, string: str) -> None:
    assert template.format(quantity=quantity) == string


def test_unrecognized_unit_format_specifier() -> None:
    with pytest.raises(ValueError):
        "{quantity::nope}".format(quantity=5 * Meter)
